import unittest

import numpy as np
from matplotlib import pyplot as plt  # DEBUGGING
import time

from chromatic_tda import plot_six_pack, SimplicialComplex  # DEBUGGING
from chromatic_tda.algorithms.radius_function import RadiusFunctionConstructor
from chromatic_tda.core.chromatic_alpha_complex_factory import CoreChromaticAlphaComplexFactory  # DEBUGGING
from chromatic_tda.core.core_chromatic_alpha_complex import CoreChromaticAlphaComplex  # DEBUGGING
from chromatic_tda.utils.timing import TimingUtils  # DEBUGGING


class RadiusFunctionTest(unittest.TestCase):
    def test_2d_2_1_1(self):
        center, radii2 = RadiusFunctionConstructor.find_smallest_circumstack(np.array([[1, 1], [-1, -1]]),
                                                                             np.array([[.2, -.1]]),
                                                                             np.array([[6, -4]]))
        radii = np.sqrt(radii2)
        assert np.isclose([2.5, -2.5], center).all()
        assert np.isclose(3.807886552931954, radii[0])
        assert np.isclose(3.3241540277189316, radii[1])
        assert np.isclose(radii[0], radii[2])

    def test_2d_2_2(self):
        c, r2 = RadiusFunctionConstructor.find_smallest_circumstack(
            np.array([[4, 6], [6, 4]]),
            np.array([[-5, 2], [-5, -2]]))
        r = np.sqrt(r2)
        assert np.isclose([0, 0], c).all()
        assert np.isclose(np.linalg.norm([4, 6]), r[0])
        assert np.isclose(np.linalg.norm([-5, 2]), r[1])

    def test_2d_2_2_1(self):
        c, r2 = RadiusFunctionConstructor.find_smallest_circumstack(
            np.array([[4, 6], [6, 4]]),
            np.array([[-5, 2], [-5, -2]]),
            np.array([[20, 25]]))
        r = np.sqrt(r2)
        assert np.isclose([0, 0], c).all()
        assert np.isclose(np.linalg.norm([4, 6]), r[0])
        assert np.isclose(np.linalg.norm([-5, 2]), r[1])
        assert np.isclose(np.linalg.norm([20, 25]), r[2])

    def test_2d_rnd_1_2_1(self):
        points = np.array([[0.64200461, 0.04874111],
                           [0.2361311, 0.26006657],
                           [0.90437823, 0.02386989],
                           [0.74969127, 0.7263785]])
        c, r2 = RadiusFunctionConstructor.find_smallest_circumstack(points[:1], points[1:3], points[3:4])
        r = np.sqrt(r2)
        assert np.isclose([0.63794903, 0.33348898], c).all()
        assert np.isclose(0.28477674882678244, r[0])
        assert np.isclose(0.4084709331753826, r[1])
        assert np.isclose(0.4084709331753826, r[2])

    def test_2d_rnd_3_1(self):
        points = np.array([[0.84230833, 0.02261223],
                           [0.99884336, 0.16265999],
                           [0.08682862, 0.20002957],
                           [1, 2]])
        c, r2 = RadiusFunctionConstructor.find_smallest_circumstack(points[:3], points[3:4])
        r = np.sqrt(r2)
        assert np.isclose([0.55590267, 0.50024081], c).all()
        assert np.isclose(0.5569176466310057, r[0])
        assert np.isclose(1.5641291689701118, r[1])

    def test_2d_rnd_3_1_case2(self):
        points = np.array([[0.84230833, 0.02261223],
                           [0.99884336, 0.16265999],
                           [0.08682862, 0.20002957],
                           [0.65710032, 0.38482416]])
        c, r2 = RadiusFunctionConstructor.find_smallest_circumstack(points[:3], points[3:4])
        r = np.sqrt(r2)
        assert np.isclose([0.55590267, 0.50024081], c).all()
        assert np.isclose(0.5569176466310057, r[0])
        assert np.isclose(0.15349907828659298, r[1])

    def test_2d_rnd_1_1_1(self):
        points = np.array([[0.31329716, 0.83100724],
                           [0.01325454, 0.80997717],
                           [0.68344234, 0.39667674]])
        c, r2 = RadiusFunctionConstructor.find_smallest_circumstack(points[:1], points[1:2], points[2:3])
        r = np.sqrt(r2)
        assert np.isclose([0.34834844, 0.60332695], c).all()
        assert np.isclose(0.23036256029675714, r[0])
        assert np.isclose(0.3936905309541866, r[1])
        assert np.isclose(0.39369052974995866, r[2])

    def test_4d_1_1(self):
        pt1, pt2 = np.array([1, 2, 3, 4]), np.array([5, 4, 5, 4])
        c, r2 = RadiusFunctionConstructor.find_smallest_circumstack(np.array([pt1]), np.array([pt2]))
        r = np.sqrt(r2)

        assert np.isclose((pt1 + pt2) / 2, c).all()
        assert np.isclose(np.linalg.norm(pt1 - pt2) / 2, r[0])
        assert np.isclose(np.linalg.norm(pt1 - pt2) / 2, r[1])


import itertools

from chromatic_tda.core.core_simplicial_complex import CoreSimplicialComplex
from chromatic_tda.core.simplicial_complex_factory import CoreSimplicialComplexFactory


class ChromaticComplexUtils:
    @staticmethod
    def get_chromatic_subcomplex(sub_complex, full_complex, relative,
                                 simplicial_complex: CoreSimplicialComplex, internal_labeling,
                                 labels_user_to_internal=None, allow_unused_labels=False) -> CoreSimplicialComplex:
        list_of_input_labels = ChromaticComplexUtils.construct_list_of_labels(
            internal_labeling=internal_labeling, labels_user_to_internal=labels_user_to_internal)
        if (full_complex is None or full_complex == '' or
                (isinstance(full_complex, str) and full_complex.lower().strip() == 'all')):
            complex_simplices = set(simplicial_complex.boundary)
        else:
            pattern = ChromaticComplexUtils.read_pattern_input(full_complex, list_of_input_labels,
                                                               check_labels=not allow_unused_labels)
            if labels_user_to_internal is not None:
                pattern = ChromaticComplexUtils.pattern_translate(pattern=pattern,
                                                                  translation_function=labels_user_to_internal)
            complex_simplices = ChromaticComplexUtils.select_simplices_with_chromatic_pattern(
                simplices=simplicial_complex.boundary.keys(), labeling_function=internal_labeling, pattern=pattern)

        if relative is None or relative == '':
            relative_simplices = set()
        elif isinstance(relative, str) and relative.lower().strip() == 'all':
            relative_simplices = set(simplicial_complex.boundary)
        else:
            pattern = ChromaticComplexUtils.read_pattern_input(relative, list_of_input_labels,
                                                               check_labels=not allow_unused_labels)
            if labels_user_to_internal is not None:
                pattern = ChromaticComplexUtils.pattern_translate(pattern=pattern,
                                                                  translation_function=labels_user_to_internal)
            relative_simplices = ChromaticComplexUtils.select_simplices_with_chromatic_pattern(
                simplices=simplicial_complex.boundary.keys(), labeling_function=internal_labeling, pattern=pattern)

        if sub_complex is None or sub_complex == '':
            sub_complex_simplices = set()
        elif isinstance(sub_complex, str) and sub_complex.lower().strip() == 'all':
            sub_complex_simplices = set(simplicial_complex.boundary)
        else:
            pattern = ChromaticComplexUtils.read_pattern_input(sub_complex, list_of_input_labels,
                                                               check_labels=not allow_unused_labels)
            if labels_user_to_internal is not None:
                pattern = ChromaticComplexUtils.pattern_translate(pattern=pattern,
                                                                  translation_function=labels_user_to_internal)
            sub_complex_simplices = ChromaticComplexUtils.select_simplices_with_chromatic_pattern(
                simplices=simplicial_complex.boundary.keys(), labeling_function=internal_labeling, pattern=pattern)

        restricted_complex: CoreSimplicialComplex = CoreSimplicialComplexFactory().create_restricted_instance(
            simplicial_complex, complex_simplices - relative_simplices)
        restricted_complex.set_sub_complex(sub_complex_simplices - relative_simplices)

        return restricted_complex

    @staticmethod
    def construct_list_of_labels(internal_labeling, labels_user_to_internal):
        if labels_user_to_internal is not None:
            return set(labels_user_to_internal.keys())
        if isinstance(internal_labeling, dict):
            return set(internal_labeling.values())
        return set(internal_labeling)

    @staticmethod
    def read_pattern_input(parameter, labels=None, check_labels=True):
        if isinstance(parameter, str):
            pattern_list_of_sets = ChromaticComplexUtils.read_pattern_input_string(parameter, labels=labels)
        else:
            pattern_list_of_sets = [set(color_set) for color_set in parameter]

        if labels is not None and check_labels:  # check that the pattern only uses the given labels
            for lab in set().union(*pattern_list_of_sets):
                if lab not in labels:
                    raise ValueError(f"There is no point labeled by `{lab}`. "
                                     f"To suppress this error, pass allow_unused_labels=True to get_complex function")

        return pattern_list_of_sets

    @staticmethod
    def read_pattern_input_string(parameter: str, labels=None):
        parameter = parameter.lower().strip()
        if parameter.endswith('chromatic'):
            chromaticity = parameter.replace('chromatic', '').replace('-', '')
            words_to_numbers = {'mono': 1, 'one': 1, 'bi': 2, 'two': 2, 'tri': 3, 'three': 3, 'tetra': 4, 'four': 4}
            if chromaticity in words_to_numbers:
                chromaticity = words_to_numbers[chromaticity]
            elif chromaticity.isnumeric() and chromaticity != '0':
                chromaticity = int(chromaticity)
            else:
                raise ValueError(f"The color pattern `{parameter}` is invalid")
            if labels is None:
                raise ValueError(
                    f"'k-chromatic' option given with labels=None. List of labels is needed for this option.")
            if labels and chromaticity > len(labels):
                chromaticity = len(labels)  # 4-chromatic subcomplex of 3-colored should still be the full complex
            pattern_list_of_sets = [set(x) for x in itertools.combinations(labels, chromaticity)]
        else:
            pattern_list_of_sets = [{int(ch) if ch.isnumeric() else ch for ch in w} for w in parameter.split(',')]

        return pattern_list_of_sets

    @staticmethod
    def simplex_satisfies_pattern(simplex, labeling_function, pattern):
        """
        Return true iff the colors of the simplex are subset of one of the patterns.
        simplex ... sorted tuple of vertices
        labeling ... list or dictionary giving a label to each possible vertex
        pattern ... collection of sets of labels
        """
        simplex_labels = {labeling_function[v] for v in simplex}
        return any(simplex_labels.issubset(s) for s in pattern)

    @staticmethod
    def select_simplices_with_chromatic_pattern(simplices, labeling_function, pattern):
        return set(simplex for simplex in simplices
                   if ChromaticComplexUtils.simplex_satisfies_pattern(simplex, labeling_function, pattern))

    @staticmethod
    def pattern_translate(pattern, translation_function):
        return [set(translation_function[lab] for lab in face if lab in translation_function) for face in pattern]

    @staticmethod
    def split_simplex_by_labels(simplex: tuple, labeling) -> dict:
        simplex_split = {}
        for v in simplex:
            lab = labeling[v]
            if lab not in simplex_split:
                simplex_split[lab] = []
            simplex_split[lab].append(v)

        return simplex_split



import numpy as np
import numpy.typing as npt
import math

from chromatic_tda.utils.timing import TimingUtils


class FloatingPointUtils:

    @staticmethod
    def is_close(a, b) -> bool:
        """
        Return True if a should be considered equal to b.
        The default method used to compare whether two floats are close.
        """
        if abs(b) > .01:
            return bool(np.isclose(a, b))
        else:
            return bool(np.isclose(a, b, atol=1e-12))

    @staticmethod
    def is_all_close(avec, bvec):
        """
        Return true if lists avec and bvec are close, i.e., if all elements are close.
        """
        if len(avec) != len(bvec):
            raise ValueError("The parameters need to be of the same length.")
        return all(FloatingPointUtils.is_close(a, b) for a, b in zip(avec, bvec))

    @staticmethod
    def is_trivial_bar(bar) -> bool:
        return FloatingPointUtils.is_close(bar[0], bar[1])

    @staticmethod
    def ensure_smaller_or_equal(value: float, *coboundary_values) -> float:
        """
        Return `value` if it's smaller than the minimum of all the other passed values or close to it.
        Otherwise, return the minimum of the other passed values
        """
        minimum_above = min(coboundary_values, default=float('inf'))
        if (FloatingPointUtils.is_close(value, minimum_above) and value != minimum_above) or value > minimum_above:
            return minimum_above
        else:
            return value

    @staticmethod
    def ensure_weights_monotonicity_and_equal_values(weight_function: dict[tuple[int, ...], float],
                                                     co_boundary: dict[tuple[int, ...], set]) -> None:
        for simplex in sorted(weight_function, key=len, reverse=True):
            weight_function[simplex] = FloatingPointUtils.ensure_smaller_or_equal(
                weight_function[simplex],
                *(weight_function[co_face] for co_face in co_boundary[simplex])
            )

    @staticmethod
    def flag_duplicates_from_reference(reference: npt.NDArray, to_check: npt.NDArray) -> npt.NDArray[bool]:
        """Return a boolean list of len(to_check) describing whether the corresponding element in to_check
        is close to some element of the reference."""
        TimingUtils().start("Flag Duplicates Among Vectors")
        flags = np.array([any(FloatingPointUtils.is_all_close(ref, x) for ref in reference) for x in to_check],
                         dtype=bool)
        TimingUtils().stop("Flag Duplicates Among Vectors")
        return flags



import itertools
from typing import Optional

import numpy as np
import numpy.typing as npt
from collections.abc import Iterable

from chromatic_tda.algorithms.chromatic_subcomplex_utils import ChromaticComplexUtils
from chromatic_tda.core.core_chromatic_alpha_complex import CoreChromaticAlphaComplex
from chromatic_tda.core.stack import StackOfSpheres
from chromatic_tda.utils.floating_point_utils import FloatingPointUtils
from chromatic_tda.utils.geometry_utils import GeometryUtils
from chromatic_tda.utils.timing import TimingUtils


class RadiusFunctionConstructor:
    @staticmethod
    def construct_sq_radius_function(alpha_complex: CoreChromaticAlphaComplex,
                                     use_morse_optimization: bool) -> dict[tuple[int, ...], float]:
        TimingUtils().start("Rad :: Construct Radius Function")

        radius_function = {}
        for dim in range(alpha_complex.simplicial_complex.dimension, 0, -1):
            simplices: set[tuple[int, ...]] = alpha_complex.simplicial_complex.dim_simplex_dict[dim]
            for simplex in simplices:
                if radius_function.get(simplex, None) is not None:
                    continue  # if radius already found at an earlier step, skip the simplex
                circumstack = RadiusFunctionConstructor.find_smallest_circumstack_of_simplex(
                    alpha_complex, simplex)
                extra_vertices = alpha_complex.simplicial_complex.get_extra_vertices_of_cofaces(simplex)
                if RadiusFunctionConstructor.is_stack_empty_of_vertices(alpha_complex, extra_vertices, circumstack):
                    radius_function[simplex] = circumstack.maximum_radius
                    if use_morse_optimization:
                        RadiusFunctionConstructor.morse_optimization_fill_in_interval(radius_function, alpha_complex,
                                                                                      simplex, circumstack)
                else:
                    co_faces = alpha_complex.simplicial_complex.co_boundary[simplex]
                    radius_function[simplex] = min(radius_function[co_face] for co_face in co_faces)
        for simplex in alpha_complex.simplicial_complex.get_simplices_of_dim(0):
            radius_function[simplex] = 0.

        TimingUtils().stop("Rad :: Construct Radius Function")
        return radius_function

    @staticmethod
    def is_stack_empty_of_vertices(alpha_complex: CoreChromaticAlphaComplex, vertices, stack: StackOfSpheres) -> bool:
        """Return True if the squared distance from center is greater or close to the corresponding color squared radius
        for all given vertices."""
        TimingUtils().start("Rad :: Check Stack Emptiness")

        for v in vertices:
            distance = np.square(alpha_complex.points[v] - stack.center).sum()
            radius = stack.radii.get(alpha_complex.internal_labeling[v], 0)
            if distance < radius and not FloatingPointUtils.is_close(distance, radius):
                return False

        TimingUtils().stop("Rad :: Check Stack Emptiness")
        return True

    @staticmethod
    def find_smallest_circumstack_of_simplex(alpha_complex: CoreChromaticAlphaComplex,
                                             simplex: tuple) -> StackOfSpheres:
        split = ChromaticComplexUtils.split_simplex_by_labels(simplex, alpha_complex.internal_labeling)
        labels, vertex_sets = zip(*split.items())
        point_sets = [alpha_complex.points[vertex_set] for vertex_set in vertex_sets]
        center, radii = RadiusFunctionConstructor.find_smallest_circumstack(*point_sets)
        circumstack = StackOfSpheres(center, {lab: rad for lab, rad in zip(labels, radii)})
        return circumstack

    @staticmethod
    def find_smallest_circumstack(*point_sets: npt.NDArray) -> tuple[npt.NDArray, npt.NDArray]:
        """For arguments B_0, ..., B_k, return the center z and radii r_0, ..., r_k defining a stack of spheres
        S_0, ..., S_k passing through B_0, ..., B_k, respectively."""
        TimingUtils().start("Rad :: Find Smallest Circumstack")
        if any(len(point_set) == 0 for point_set in point_sets):
            raise ValueError("Point sets need to be non-empty.")
        center, radii2 = RadiusFunctionConstructor.find_smallest_circumstack_weighted_circumspheres(*point_sets)
        TimingUtils().stop("Rad :: Find Smallest Circumstack")
        return center, radii2

    # @staticmethod
    # def find_smallest_circumstack_center_miniball(*point_sets: npt.NDArray) -> npt.NDArray:
    #     TimingUtils().start("Rad :: Find Smallest Circumstack :: Miniball")
    #     lengths = [len(point_set) for point_set in point_sets]
    #     equispace = GeometryUtils.construct_equispace(*point_sets)
    #     points = np.concatenate(point_sets)
    #     points_reflected = GeometryUtils.reflect_points_through_affine_space(equispace, *points)
    #     points_reflected_filtered = np.concatenate([
    #         refl[~FloatingPointUtils.flag_duplicates_from_reference(orig, refl)]
    #         for orig, refl in zip(point_sets, RadiusFunctionConstructor.split_points(points_reflected, lengths))
    #     ])  # filter away duplicates
    #     center, rad = MiniballAlgorithm.find_smallest_enclosing_ball(
    #         np.concatenate([points, points_reflected_filtered]))
    #     TimingUtils().stop("Rad :: Find Smallest Circumstack :: Miniball")
    #
    #     return center

    @staticmethod
    def find_smallest_circumstack_weighted_circumspheres(*point_sets: npt.NDArray) -> tuple[npt.NDArray, npt.NDArray]:
        TimingUtils().start("Rad :: Find Smallest Circumstack :: Weighted Circumspheres")
        equispace = GeometryUtils.construct_equispace(*point_sets)

        points = np.array([point_set[0] for point_set in point_sets])  # each point set projects to a single point

        if equispace.dimension < 0:
            center = equispace.shift
            radii2 = np.square(points - center).sum(axis=1)
            return center, radii2

        points_projected = GeometryUtils.orthogonal_projection_onto_affine_space(equispace, *points)
        weights = np.square(points - points_projected).sum(axis=1)
        candidate = (np.array([]), np.array([]), float('inf'))  # center, radii, max_radius
        for k in range(1, equispace.dimension + 2):  # if dim = 2, we want up to 3 points, and we need range(1, 4)
            for choice in itertools.combinations(range(len(point_sets)), k):
                center, _ = GeometryUtils.circumsphere_of_weighted_points(points_projected[list(choice)],
                                                                          weights[list(choice)])
                radii2 = np.square(points - center).sum(axis=1)
                rad2 = max(radii2)
                if rad2 < candidate[2]:
                    candidate = (center, radii2, rad2)

        TimingUtils().stop("Rad :: Find Smallest Circumstack :: Weighted Circumspheres")

        return candidate[0], candidate[1]

    @staticmethod
    def split_points(points: npt.NDArray, lengths) -> list[npt.NDArray, ...]:
        TimingUtils().start("Rad :: Split Points Into Colors")

        splits = np.zeros(len(lengths), dtype=int)
        for i, ln in enumerate(lengths):
            splits[i] = splits[i - 1] + ln

        TimingUtils().stop("Rad :: Split Points Into Colors")
        return np.split(points, splits)[:-1]

    @staticmethod
    def compute_kkt_solution(points: Iterable[npt.NDArray], labels: Iterable, circumstack: StackOfSpheres) \
            -> Optional[npt.NDArray]:
        """Return vector of lambdas ordered as simplex if valid solution exists, otherwise return None.
        WARNING: `circumstack` is assumed to be a circumstack of `points`, `labels`, condition not checked
        WARNING: Emptiness is *not* checked.
        WARNING: `points` is assumed to be all points that `circumstack` goes through --- if not the case,
                 solution might not exist, and None is returned"""
        TimingUtils().start("Rad :: Morse :: Compute KKT Solution")
        labels_to_int = {lab: i for i, lab in enumerate(circumstack.radii.keys())}
        number_of_labels = len(labels_to_int)
        point_part = np.array([
            np.concatenate([np.eye(1, number_of_labels + 1, k=labels_to_int[lab]).flatten(), pt])
            for pt, lab in zip(points, labels)
        ])
        color_part = np.zeros(shape=(len(circumstack.maximum_labels), point_part.shape[1]))
        for i, lab in enumerate(circumstack.maximum_labels):
            color_part[i, labels_to_int[lab]] = -1
            color_part[i, number_of_labels] = 1
        a_mat = np.concatenate([point_part, color_part]).transpose()
        b_vec = np.concatenate([np.eye(1, number_of_labels + 1, k=number_of_labels).flatten(), circumstack.center])

        x, res, rk, s = np.linalg.lstsq(a_mat, b_vec, rcond=None)

        TimingUtils().stop("Rad :: Morse :: Compute KKT Solution")
        if len(res) > 0 and not FloatingPointUtils.is_close(res[0], 0):
            return None
        else:
            return x[:len(point_part)]

    @staticmethod
    def generate_radius_function_interval(simplex: tuple, coefficients: Iterable) -> list[tuple, ...]:
        """Return a list of simplices [min, simplex] - simplex, where min are all the vertices for which
        the corresponding coefficient is strictly positive."""
        TimingUtils().start("Rad :: Morse :: Generate Interval Simplices")
        minimal_simplex = {v for v, coef in zip(simplex, coefficients) if coef > 0}
        non_positive_vertices = set(simplex) - minimal_simplex
        interval = [tuple(sorted(minimal_simplex.union(extra)))
                    for k in range(len(non_positive_vertices))
                    for extra in itertools.combinations(non_positive_vertices, k)]
        TimingUtils().stop("Rad :: Morse :: Generate Interval Simplices")
        return interval

    @staticmethod
    def morse_optimization_fill_in_interval(radius_function: dict[tuple, float],
                                            alpha_complex: CoreChromaticAlphaComplex,
                                            simplex: tuple,
                                            circumstack: StackOfSpheres) -> None:
        points, labels = zip(*[(alpha_complex.points[v], alpha_complex.internal_labeling[v])
                               for v in simplex])
        lambdas = RadiusFunctionConstructor.compute_kkt_solution(points, labels, circumstack)
        if lambdas is not None:
            interval = RadiusFunctionConstructor.generate_radius_function_interval(simplex, lambdas)
            for interval_simplex in interval:
                radius_function[interval_simplex] = circumstack.maximum_radius



import unittest
from pathlib import Path
import json
import numpy as np

from chromatic_tda.core.chromatic_alpha_complex_factory import CoreChromaticAlphaComplexFactory
from chromatic_tda.core.core_chromatic_alpha_complex import CoreChromaticAlphaComplex


class ChromaticAlphaComplexConstructionTest(unittest.TestCase):

    data_folder = Path(__file__).parent / 'test_data'

    def test_chromatic_alpha_random_10_10(self):
        result = self.single_test(data_name='chralph_random_10_10')
        assert result

    def test_chromatic_alpha_random_10_10_5(self):
        result = self.single_test(data_name='chralph_random_10_10_5')
        assert result

    def test_chromatic_alpha_random_5_2_1(self):
        result = self.single_test(data_name='chralph_random_5_2_1')
        assert result

    def test_chromatic_alpha_random_15_15(self):
        result = self.single_test(data_name='chralph_random_15_15')
        assert result

    def test_chromatic_alpha_random_15_8_7(self):
        result = self.single_test(data_name='chralph_random_15_8_7')
        assert result

    def test_chromatic_alpha_integer_2_2(self):
        result = self.single_test(data_name='chralph_integer_2_2')
        assert result

    def single_test(self, data_name, data_folder=None):
        data = self.load_data(data_name, data_folder)
        factory = CoreChromaticAlphaComplexFactory(points=data['points'], labels=data['labels'])
        alpha = factory.create_instance(lift_perturbation=1e-9, point_perturbation=None)
        return self.compare_complex(alpha, data['weight_function'])

    def load_data(self, data_name, data_folder):
        if data_folder is None:
            data_folder = self.data_folder
        with open(data_folder / (data_name + '.json'), 'r') as file:
            data = json.load(file)
        return data

    @staticmethod
    def compare_complex(alpha: CoreChromaticAlphaComplex, reference_weight_function):
        computed_simplices = set(alpha.simplicial_complex.get_simplices())
        reference_simplices = set(tuple(sorted(item['simplex'])) for item in reference_weight_function)
        if computed_simplices != reference_simplices:
            return False
        for item in reference_weight_function:
            computed_weight = alpha.simplicial_complex.get_simplex_weight(tuple(sorted(item['simplex'])))
            reference_weight = item['weight']
            if not np.isclose(computed_weight, reference_weight):
                return False
        return True  # simplices identical to the reference, and all the weights are close to the reference



import numpy as np

from chromatic_tda.algorithms.chromatic_subcomplex_utils import ChromaticComplexUtils
from chromatic_tda.algorithms.persistence_algorithm import PersistenceAlgorithm
from chromatic_tda.core.core_simplicial_complex import CoreSimplicialComplex
from chromatic_tda.core.simplicial_complex_factory import CoreSimplicialComplexFactory


class SimplicialComplex:

    GROUPS = ['kernel', 'sub_complex', 'image', 'complex', 'cokernel', 'relative']
    core_complex: CoreSimplicialComplex

    def __init__(self, data):
        if isinstance(data, CoreSimplicialComplex):
            self.core_complex = data
        else:
            self.core_complex = CoreSimplicialComplexFactory().create_instance(data)

    def __iter__(self):
        yield from self.core_complex

    def __len__(self) -> int:
        return len(self.core_complex)

    def __contains__(self, element) -> bool:
        return element in self.core_complex

    def compute_persistence(self):
        PersistenceAlgorithm(simplicial_complex=self.core_complex).compute_persistence()

    def dimension(self) -> int:
        """Return the dimension of the simplicial complex"""
        return self.core_complex.get_dimension()

    def bars(self, group : str, dim : int = None, only_finite : bool = False, return_as : str = 'dict'):
        """Return persistence bars for a given group as. If dimension is given, bars are returned as a list of
         (birth, death) pairs. If no dimension is given, all dimensions are returned in one of the following formats:
            - dict ... a dictionary {dimension : list_of_bars}
            - list ... a list of tuples (dimension, (birth, death))
        Default return format is `dict`.

        Keyword arguments:
            dim ... Dimension. If an integer is given, returns a list of bars of only that dimension (default: None).
            only_finite ... If True, only return the finite bars (default: False).
            return_as ... Either 'dict' or 'list'. Switch return format as described above. (default: 'dict')

        Remark: If compute_persistence was not called on the SimplicialComplex yet, it is called.
        """
        if group not in self.GROUPS:
            raise ValueError(f'The group needs to be one of the following: ' + ', '.join(self.GROUPS))
        if len(self.core_complex.birth_death) == 0:
            self.compute_persistence()
        bars = self.core_complex.get_bars_list(group, dim=dim, only_finite=only_finite)
        if dim is not None or return_as == 'list':
            return bars
        elif return_as == 'dict':
            bars_dict = {}
            for dim, bar in bars:
                if dim not in bars_dict:
                    bars_dict[dim] = []
                bars_dict[dim].append(bar)
            return bars_dict
        else:
            raise ValueError('The argument `return_as` can only be "dict" or "list".')

    def bars_six_pack(self, only_finite : bool = False, return_as : str = 'dict') -> dict:
        """Return the whole six-pack as a dictionary of the form { group : bars_of_the_group }. Bars of one group
        are returned in one of the following formats:
            - dict ... a dictionary {dimension : list_of_bars}
            - list ... a list of tuples (dimension, (birth, death))
        Default return format is `dict`.

        Keyword arguments:
            only_finite ... If True, only return the finite bars (default: False).
            return_as ... Either 'dict' or 'list'. Switch return format as described above. (default: 'dict')

        Remark: If compute_persistence was not called on the SimplicialComplex yet, it is called.
        """
        return {group : self.bars(group, dim=None, only_finite=only_finite, return_as=return_as)
                for group in self.GROUPS}

    def bars_finite_1norm(self, group : str, dim : int):
        """Return the 1-norm of finite bars of the given group and dimension"""
        return sum([death - birth for birth, death in self.bars(group, dim=dim, only_finite=True)])

    def diagrams_list(self, group_dimension_pattern, only_finite=False):
        """Return list of persistence diagrams defined by the `group_dimension_pattern`. The pattern is a list of
        pairs (group, dimension). The method respects the order given.
        """
        return [self.bars(group, dim, only_finite=only_finite) for group, dim in group_dimension_pattern]

    # def get_1norm_share(self, group_numerator : str, group_denominator : str, dim : int):
    #     """Return the ratio of finite bars 1-norms of group_numerator and group_denominator"""
    #     numerator = self.bars_finite_1norm(group_numerator, dim)
    #     denominator = self.bars_finite_1norm(group_denominator, dim)
    #     return numerator/denominator if denominator else 0
    
    def simplices(self) -> list:
        """Return list of all simplices sorted by dimension and then lexicographically (w.r.t. vertex indices)."""
        return self.core_complex.get_simplices()

    def weight_function(self, simplex=None):
        """If simplex is given, return the weight/radius of the simplex.
        If no simplex is given, return the weight/radius function as a dictionary {simplex : weight}."""
        if simplex is None:
            return self.core_complex.get_weight_function_copy()
        return self.core_complex.get_simplex_weight(simplex)

    def simplices_of_dim(self, dim : int) -> list:
        """Return list of all simplices ."""
        return self.core_complex.get_simplices_of_dim(dim)

    def simplices_sub_complex(self) -> list:
        """Return list of all simplices of the sub_complex, sorted by dimension and then
        lexicographically (w.r.t. vertex indices)."""
        return self.core_complex.get_sub_complex_simplices()

    def is_in_sub_complex(self, simplex) -> bool:
        """Return True iff the given simplex is in the sub-complex."""
        return simplex in self.core_complex.sub_complex

    def set_simplex_weights(self, weight_function : dict, default_value : float = 0):
        """Set the weights of simplices. The `weight_function` is a dictionary {simplex : weight}.
        The `default_value` is used for all simplices not present in `weight_function`.
        USE WITH CAUTION: The method is meant for creating a new simplicial complex with your own filtration. If you
        change this, you need to manually run `compute_persistence` again."""
        self.core_complex.set_simplex_weights(weight_function, default_value)

    def set_sub_complex(self, simplices):
        """Set the sub-complex by passing its maximal (generating) simplices.
        USE WITH CAUTION: The method is meant for creating a new simplicial complex with your own filtration. If you
        change this, you need to manually run `compute_persistence` again."""
        self.core_complex.set_sub_complex(simplices)

    def get_chromatic_subcomplex(self, labeling, sub_complex=None, full_complex=None, relative=None,
                                 allow_unused_labels=False):
        """Ignoring the current subcomplex, return a new SimplicialComplex as a chromatic subcomplex given by
        the parameters (see docstring of ChromaticAlphaComplex.get_simplicial_complex).
        Argument labeling is a dictionary or a list such that labeling[vertex] = label;
        A previously defined sub_complex of the complex is NOT preserved."""
        return SimplicialComplex(ChromaticComplexUtils.get_chromatic_subcomplex(
            internal_labeling=labeling, sub_complex=sub_complex, full_complex=full_complex, relative=relative,
            simplicial_complex=self.core_complex, allow_unused_labels=allow_unused_labels)
        )

    def chain_boundary(self, simplices) -> set:
        """Return boundary of a chain of simplices, i.e., the symmetric difference of boundaries of the given collection
         of simplices"""
        return self.core_complex.chain_boundary(simplices)


from chromatic_tda.algorithms.reduce_matrix import MatrixReduction
from chromatic_tda.core.core_simplicial_complex import CoreSimplicialComplex
from chromatic_tda.utils.filter_functions import FilterFunctions


class PersistenceAlgorithm:  # why is this not a singleton? with complex as a parameter for each function?
    complex: CoreSimplicialComplex
    filters: FilterFunctions

    def __init__(self, simplicial_complex: CoreSimplicialComplex) -> None:
        if set(simplicial_complex.simplex_weights.keys()) != set(simplicial_complex.boundary.keys()):
            raise ValueError("The weight function does not match the simplices of the simplicial complex.")
        self.complex = simplicial_complex
        self.filters = FilterFunctions(simplicial_complex.simplex_weights, simplicial_complex.sub_complex)

    def compute_persistence(self) -> None:
        """
        Compute the following persistence homology:
            complex -- PH of the whole complex
            sub_complex -- PH only of the sub_complex
            kernel -- kernel PH
            image  -- image PH
            cokernel -- co_kernel PH
        """
        self._compute_persistence_data()
        self.complex.birth_death = {}
        self._compute_birth_death_complex()
        self._compute_birth_death_sub_complex()
        self._compute_birth_death_image()
        self._compute_birth_death_kernel()
        self._compute_birth_death_cokernel()
        self._compute_persistence_relative()

    def _compute_birth_death_complex(self) -> None:
        Rf = self.complex.persistence_data['complex']['reduced_matrix']
        low_inv_f = self.complex.persistence_data['complex']['pivots']

        birth = set(s for s in Rf if len(Rf[s]) == 0)
        death = set(s for s in Rf if len(Rf[s]) != 0)
        pairs = set((k, v) for k, v in low_inv_f.items())
        killed = set(pair[0] for pair in pairs)
        essential = birth - killed

        self.complex.birth_death['complex'] = {
            'birth': birth,
            'death': death,
            'essential': essential,
            'pairs': pairs
        }

    def _compute_birth_death_sub_complex(self) -> None:
        Rg = self.complex.persistence_data['sub_complex']['reduced_matrix']
        low_inv_g = self.complex.persistence_data['sub_complex']['pivots']

        birth = set(s for s in Rg if len(Rg[s]) == 0)
        death = set(s for s in Rg if len(Rg[s]) != 0)
        pairs = set((k, v) for k, v in low_inv_g.items())
        killed = set(pair[0] for pair in pairs)
        essential = birth - killed

        self.complex.birth_death['sub_complex'] = {
            'birth': birth,
            'death': death,
            'essential': essential,
            'pairs': pairs
        }

    def _compute_birth_death_image(self) -> None:
        Rg = self.complex.persistence_data['sub_complex']['reduced_matrix']
        low_inv_im = self.complex.persistence_data['image']['pivots']

        birth = set(s for s in Rg if len(Rg[s]) == 0)
        pairs = set((k, v) for k, v in low_inv_im.items() if k in self.complex.sub_complex)
        death = set(pair[1] for pair in pairs)
        killed = set(pair[0] for pair in pairs)
        essential = birth - killed

        self.complex.birth_death['image'] = {
            'birth': birth,
            'death': death,
            'essential': essential,
            'pairs': pairs
        }

    def _compute_birth_death_kernel(self) -> None:
        Rf = self.complex.persistence_data['complex']['reduced_matrix']
        Rg = self.complex.persistence_data['sub_complex']['reduced_matrix']
        low_inv_im = self.complex.persistence_data['image']['pivots']
        low_inv_ker = self.complex.persistence_data['kernel']['pivots']

        birth = set(v for k, v in low_inv_im.items() if (v not in self.complex.sub_complex and
                                                         k in self.complex.sub_complex))
        pairs = set((k, v) for k, v in low_inv_ker.items() if (v in self.complex.sub_complex and  # tau in L
                                                               len(Rg[v]) != 0 and  # tau negative in Rg
                                                               len(Rf[v]) == 0))  # tau positive in Rf
        death = set(pair[1] for pair in pairs)
        killed = set(pair[0] for pair in pairs)
        essential = birth - killed

        self.complex.birth_death['kernel'] = {
            'birth': birth,
            'death': death,
            'essential': essential,
            'pairs': pairs
        }

    def _compute_birth_death_cokernel(self) -> None:
        Rim = self.complex.persistence_data['image']['reduced_matrix']
        low_im = {v: k for k, v in self.complex.persistence_data['image']['pivots'].items()}
        Rg = self.complex.persistence_data['sub_complex']['reduced_matrix']
        low_inv_cok = self.complex.persistence_data['cokernel']['pivots']

        birth = set(s for s in Rim if len(Rim[s]) == 0 and (s not in self.complex.sub_complex or len(Rg[s]) != 0))
        pairs = set((k, v) for k, v in low_inv_cok.items()
                    if len(Rim[v]) != 0 and low_im[v] not in self.complex.sub_complex)
        death = set(pair[1] for pair in pairs)
        killed = set(pair[0] for pair in pairs)
        essential = birth - killed
        self.complex.birth_death['cokernel'] = {
            'birth': birth,
            'death': death,
            'essential': essential,
            'pairs': pairs
        }

    def _compute_persistence_relative(self) -> None:
        R = self.complex.persistence_data['relative']['reduced_matrix']
        low_inv = self.complex.persistence_data['relative']['pivots']

        birth = set(s for s in R if len(R[s]) == 0)
        death = set(s for s in R if len(R[s]) != 0)
        pairs = set((k, v) for k, v in low_inv.items())
        killed = set(pair[0] for pair in pairs)
        essential = birth - killed

        self.complex.birth_death['relative'] = {
            'birth': birth,
            'death': death,
            'essential': essential,
            'pairs': pairs
        }

    def _compute_persistence_data(self) -> None:
        self.complex.persistence_data = {}

        # TODO (low priority): Only compute what is needed. The prerequisites graph is as follows:
        #   complex    -R--> image
        #   complex    -R--> relative
        #   complex    -RV-> kernel
        #   complex    -R--> co_kernel
        #   sub_complex -RV-> co_kernel
        # Now user can give a list of desired groups, we add prerequisites to the list
        # and then only compute the data from the list. ! Also need to change compute_persistence !

        self.complex.persistence_data['complex'] = MatrixReduction.reduce(
            self.complex.boundary,
            order_function=self.filters.filter_function_rad(),
            return_reduction_matrix=True)

        self.complex.persistence_data['sub_complex'] = MatrixReduction.reduce(
            {simplex: self.complex.boundary[simplex] for simplex in self.complex.sub_complex},  # bnd mat of subcomplex
            order_function=self.filters.filter_function_rad(),
            return_reduction_matrix=True)

        self.complex.persistence_data['image'] = MatrixReduction.reduce(
            self.complex.persistence_data['complex']['reduced_matrix'],  # instead of (self.complex.boundary) for performance
            order_function=self.filters.filter_function_rad(),
            order_function_row=self.filters.filter_function_rad_sub_first(),
            return_reduction_matrix=False)

        R = self.complex.persistence_data['complex']['reduced_matrix']  # should be R_im, but coincides on cycles with R_f
        V = self.complex.persistence_data['complex']['reduction_matrix']  # should be V_im, but coincides on cycles with V_f
        self.complex.persistence_data['kernel'] = MatrixReduction.reduce(
            {simplex: V[simplex] for simplex in V if len(R[simplex]) == 0},  # columns of V which represent cycles
            order_function=self.filters.filter_function_rad(),
            order_function_row=self.filters.filter_function_rad_sub_first(),
            return_reduction_matrix=False)

        Vg = self.complex.persistence_data['sub_complex']['reduction_matrix']
        Rg = self.complex.persistence_data['sub_complex']['reduced_matrix']
        Df = self.complex.persistence_data['complex']['reduced_matrix']  # We can take R rather than D, because we only replace
                                                                         # cycle columns, so all reductions are still valid.
        D_cok = {simplex: Vg[simplex] if (simplex in Rg and len(Rg[simplex]) == 0) else Df[simplex] for simplex in Df}
        self.complex.persistence_data['cokernel'] = MatrixReduction.reduce(
            D_cok,
            order_function=self.filters.filter_function_rad(),
            return_reduction_matrix=False)

        matrix_relative = {
            s: {t for t in self.complex.persistence_data['complex']['reduced_matrix'][s] if t not in self.complex.sub_complex}
            for s in self.complex.persistence_data['complex']['reduced_matrix'] if s not in self.complex.sub_complex}
        self.complex.persistence_data['relative'] = MatrixReduction.reduce(
            matrix_relative,
            order_function=self.filters.filter_function_rad(),
            return_reduction_matrix=False)

    @staticmethod
    def get_birth_death_from_matrix(R, low_inv) -> dict:
        """Given a matrix R a low_inv function,
        return birth, death and essential simplices."""
        birth = set(s for s in R if len(R[s]) == 0)
        death = set(s for s in R if len(R[s]) != 0)
        essential = set(s for s in birth if s not in low_inv.keys())

        return {
            'birth': birth,
            'death': death,
            'essential': essential
        }



import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes

from chromatic_tda.entities.simplicial_complex import SimplicialComplex
from chromatic_tda.utils.floating_point_utils import FloatingPointUtils


def plot_persistence_diagram(bars, ax=None, **kwargs):
    """Plot a single persistence diagram. Bars can be given in several different formats:
        - list of bars, i.e., (birth, death) tuples
        - list of (dimension, (birth, death)) tuples ... prints each dimension with different marker/color
        - dictionary of the form { dimension : list_of_bars }
    A matplotlib axis can be given as the ax argument.
    If keyword argument 'max_dimension' is a number, only print diagrams up to that dimension.
    Various options can be given via keyword arguments, many coming from matplotlib options. Include the following,
    with default values in parentheses:
        size (5), facecolor ('white'), aspect ('equal'), 'lim' (None), markersize (6), fillstyle ('none'), alpha (.7)
        ticks_fontsize (9), label_fontsize (11), title_fontsize (17), legend_fontsize (11)
        axes_labels (True), only_finite (False), label_inf (True), infinite_line_legend (True),
        color ({0: 'tab:green', 1: 'blue', 2: 'black', 3: 'orange'}), marker ({0: 'D', 1: 'o', 2: 's', 3: '*'}),
    """
    bars_dict = PlottingUtils().process_input_to_dict_form(bars)
    if 'max_dimension' in kwargs:
        bars_dict = {dim : bars for dim, bars in bars_dict.items() if dim <= kwargs['max_dimension']}
    bars_dict_fin, bars_dict_inf = PlottingUtils().split_finite_infinite_dictionary(bars_dict)

    if 'lim' in kwargs and kwargs['lim'] is not None:
        xlim, ylim = PlottingUtils().parse_plot_limit_argument(kwargs['lim'])
    else:  # is not in kwargs (or is None)
        xlim, ylim = PlottingUtils().find_plot_limits(bars_dict_fin, **kwargs)

    infinity_position = kwargs.get('infinity_position', ylim[1] * .98)

    if ax:
        plt.sca(ax)
    else:
        plt.figure(figsize=(kwargs.get('size', 5), kwargs.get('size', 5)),
                   facecolor=kwargs.get('facecolor', 'white'))
    if not FloatingPointUtils.is_close(*xlim):  # to avoid passing identical low and high xlim to pyplot
        plt.xlim(xlim)
    if not FloatingPointUtils.is_close(*ylim):  # to avoid passing identical low and high ylim to pyplot
        plt.ylim(ylim)
    plt.gca().set_aspect(kwargs.get('aspect', 'equal'))
    plt.xticks(fontsize=kwargs.get('ticks_fontsize', 9))
    plt.yticks(fontsize=kwargs.get('ticks_fontsize', 9))
    if kwargs.get('axes_labels', True):
        plt.xlabel('birth', fontsize=kwargs.get('label_fontsize', 11))
        plt.ylabel('death', fontsize=kwargs.get('label_fontsize', 11))

    color = kwargs.get('color', {0: 'tab:green', 1: 'blue', 2: 'black', 3: 'orange'})
    marker = kwargs.get('marker', {0: 'D', 1: 'o', 2: 's', 3: '*'})
    if type(color) == list:
        color = {d: color[i] for i, d in enumerate(sorted(bars_dict)) if len(color) > i}
    elif type(color) == str:
        color = {d: color for d in bars_dict}

    if type(marker) == list:
        marker = {d: marker[i] for i, d in enumerate(sorted(bars_dict)) if len(marker) > i}
    elif type(marker) == str:
        marker = {d: marker for d in bars_dict}

    plt.axvline(0, color='gray', linewidth=1, alpha=.5)
    plt.plot(xlim, xlim, color='black', linewidth=1, alpha=.5)
    if 'title' in kwargs:
        plt.title(kwargs['title'], fontsize=kwargs.get('title_fontsize', 17))
    if not kwargs.get('only_finite', False) and any(len(bars_inf) > 0 for bars_inf in bars_dict_inf.values()):  # if also plotting infinite
        label = 'inf' if kwargs.get('label_inf', True) else None
        plt.axhline(infinity_position, color='tab:red', alpha=.5, linewidth=1, label=label)
        if label is not None:
            kwargs['infinite_line_legend'] = True
    for d in sorted(bars_dict_fin):
        if kwargs.get('only_finite', False):
            bars = bars_dict_fin[d]
        else:
            bars = list(bars_dict_fin[d]) + [[birth, infinity_position] for birth, death in bars_dict_inf[d]]
        plt.plot(
            *zip(*bars),
            linestyle='none',
            markersize=kwargs.get('markersize', 6),
            marker=marker.get(d, 'o'),
            color=color.get(d, None),
            fillstyle=kwargs.get('fillstyle', 'none'),
            alpha=kwargs.get('alpha', .7),
            label=str(d) if d != '' else None
        )
    if len(set(bars_dict_fin.keys()) - {''}) > 0 or kwargs.get('infinite_line_legend', False):
        plt.legend(loc='center right', fontsize=kwargs.get('legend_fontsize', 11))

    return plt.gcf(), plt.gca()


def plot_six_pack(data, axs=None, **kwargs):
    """Plot a six-pack of persistence diagrams. Data can be given in several different formats:
        - SimplicialComplex object
        - Dictionary with keys 'kernel', 'relative', 'cokernel', 'sub_complex', 'image', 'complex', and values
          the bars for the given group. The bars can be given in the following formats:
            - list of bars, i.e., (birth, death) tuples
            - list of (dimension, (birth, death)) tuples ... prints each dimension with different marker/color
            - dictionary of the form { dimension : list_of_bars }
    A list of six matplotlib axes can be given as the axs argument.
    Various options can be given via keyword arguments, many coming from matplotlib options. For more details see
    docstring of plot_persistence_diagram.
    If keyword argument 'max_dimension' is a number, only print diagrams up to that dimension.
    """
    if isinstance(data, SimplicialComplex):
        data = data.bars_six_pack()
    if isinstance(data, dict):
        data = {group : PlottingUtils().process_input_to_dict_form(bars) for group, bars in data.items()}
    else:
        raise TypeError("The data to plot a six-pack need to be given either as SimplicialComplex or a dictionary.")

    if 'max_dimension' in kwargs:
        for group in data:
            data[group] = {dim : bars for dim, bars in data[group].items() if dim <= kwargs['max_dimension']}

    if axs is None:
        fig, axs = plt.subplots(2, 3,
                                figsize=(3 * kwargs.get('size', 5), 2 * kwargs.get('size', 5)),
                                facecolor=kwargs.get('facecolor', 'white'))
    axs = np.array(axs).flatten()
    if 'lim' in kwargs and kwargs['lim'] is not None:
        xlim, ylim = PlottingUtils().parse_plot_limit_argument(kwargs['lim'])
    else:
        xlims, ylims = zip(*[PlottingUtils().find_plot_limits(bars_dict) for group, bars_dict in data.items()])
        xlim = (min(x[0] for x in xlims), max(x[1] for x in xlims))
        ylim = (min(y[0] for y in ylims), max(y[1] for y in ylims))
    kwargs['lim'] = (xlim, ylim)

    if kwargs.get('axes_labels', True):
        plt.gcf().supxlabel('birth', fontsize=kwargs.get('label_fontsize', 11), y=.06)
        plt.gcf().supylabel('death', fontsize=kwargs.get('label_fontsize', 11), x=.09)
        kwargs['axes_labels'] = False

    groups_order = ['kernel', 'relative', 'cokernel', 'sub_complex', 'image', 'complex']
    if kwargs.get('allow_arbitrary_keys', False):
        groups = sorted(data.keys(), key=lambda k: groups_order.index(k) if k in groups_order else len(groups_order))
    else:
        groups = groups_order
        if set(groups) != set(data.keys()):
            raise ValueError("Keys in the data dictionary need to be 'kernel', 'relative', 'cokernel', 'sub_complex', "
                             "'image', 'complex'. To allow different keys, pass allow_arbitrary_keys=True.")

    for group, ax in zip(groups, axs):
        plot_persistence_diagram(data[group], ax=ax, title=group.replace('_', '-'), **kwargs)

    return plt.gcf(), axs


def plot_labeled_point_set(points, labels, ax=None, **kwargs):
    """A simple plot of points and labels.
    Accepted keyword arguments: size, facecolor, markers, colors, alpha
    Returns
    -------
    Figure, Axes"""
    if ax:
        plt.sca(ax)
    else:
        plt.figure(figsize=(kwargs.get('size', 7), kwargs.get('size', 7)),
                   facecolor=kwargs.get('facecolor', 'white'))
    if len(points) > 0 and len(points[0]) != 2:
        raise ValueError('Only two-dimensional point sets can be plotted.')
    colors = kwargs.get('colors', {})
    markers = kwargs.get('markers', {lab: marker for lab, marker in zip(sorted(set(labels)), ('o', 's', '^'))})
    for current_label in sorted(set(labels)):
        plt.plot(*zip(*[pt for pt, lab in zip(points, labels) if lab == current_label]),
                 linestyle='none',
                 marker=markers.get(current_label, 'o'),
                 color=colors.get(current_label, None),
                 label=str(current_label),
                 alpha=kwargs.get('alpha', 1))
    plt.gca().set_aspect('equal')
    if kwargs.get('legend', False):
        plt.legend()
    return plt.gcf(), plt.gca()

def plot_edges(points, edges, ax=None, **kwargs):
    """Plot the given set of edges with vertex position given by points.
    Parameters
    ----------
    points ... list of points in a plane
    edges  ... list of abstract edges, e.g., an edge (1, 5) will connect points[1] and points[5]
    ax     ... axis to witch the edges are drawn, e.g. the second return value of plot_labeled_point_set

    Returns
    -------
    Figure, Axes
    """
    kwargs['facecolor'] = kwargs.get('facecolor', 'white')
    kwargs['size'] = kwargs.get('size', 7)
    kwargs['color'] = kwargs.get('color', '#000072')
    kwargs['linewidth'] = kwargs.get('linewidth', 1)
    if ax:
        plt.sca(ax)
    else:
        plt.figure(figsize=(kwargs['size'], kwargs['size']),
                   facecolor=kwargs['facecolor'])
    kwargs.pop('facecolor')
    kwargs.pop('size')
    for edge in edges:
        edge_points = [points[i] for i in edge]
        plt.plot(*zip(*edge_points), **kwargs)

    return plt.gcf(), plt.gca()



class PlottingUtils:

    X_AXIS_EXTRA_RELATIVE_SPACE = 0.03
    Y_AXIS_EXTRA_RELATIVE_SPACE = 0.02
    Y_AXIS_EXTRA_RELATIVE_SPACE_WITH_INFINITY = 0.07

    def __init__(self):
        pass

    @staticmethod
    def find_max_birth_single_diagram(bars) -> float:
        return max((birth for birth, death in bars), default=0)

    @staticmethod
    def find_max_birth_dictionary(bars_dict) -> float:
        return max((PlottingUtils.find_max_birth_single_diagram(bars) for bars in bars_dict.values()), default=0)

    @staticmethod
    def find_max_finite_death_single_diagram(bars) -> float:
        return max((death for birth, death in bars if death < float('inf')), default=0)

    @staticmethod
    def find_max_finite_death_dictionary(bars_dict) -> float:
        return max((PlottingUtils.find_max_finite_death_single_diagram(bars) for bars in bars_dict.values()), default=0)

    @staticmethod
    def find_plot_limits(bars_dict, **kwargs):
        maxdeath = PlottingUtils.find_max_finite_death_dictionary(bars_dict)
        maxbirthdeath = max(maxdeath, PlottingUtils.find_max_birth_dictionary(bars_dict))
        if kwargs.get('only_finite', False):
            lim_left = -PlottingUtils.X_AXIS_EXTRA_RELATIVE_SPACE * maxdeath
            xlim = (lim_left, maxdeath * (1 + PlottingUtils.Y_AXIS_EXTRA_RELATIVE_SPACE))
            ylim = (lim_left, maxdeath * (1 + PlottingUtils.Y_AXIS_EXTRA_RELATIVE_SPACE))
        else:
            lim_left = -PlottingUtils.X_AXIS_EXTRA_RELATIVE_SPACE * maxbirthdeath
            xlim = (lim_left, maxbirthdeath * (1 + PlottingUtils.Y_AXIS_EXTRA_RELATIVE_SPACE))
            ylim = (lim_left, maxbirthdeath * (1 + PlottingUtils.Y_AXIS_EXTRA_RELATIVE_SPACE_WITH_INFINITY))

        return xlim, ylim

    @staticmethod
    def parse_plot_limit_argument(lim) -> tuple[tuple[float, float], tuple[float, float]]:
        if isinstance(lim, list) or isinstance(lim, tuple):
            if isinstance(lim[0], list) or isinstance(lim[0], tuple):
                xlim = lim[0]
                ylim = lim[1]
            else:  # is (xmax, ymax) tuple
                lim_left = -PlottingUtils.X_AXIS_EXTRA_RELATIVE_SPACE * max(lim[0], lim[1])
                xlim = (lim_left, lim[0])
                ylim = (lim_left, lim[1])
        else:  # is just one number
            lim_left = -PlottingUtils.X_AXIS_EXTRA_RELATIVE_SPACE * lim
            xlim = (lim_left, lim)
            ylim = (lim_left, lim)

        return xlim, ylim

    @staticmethod
    def convert_tuple_form_to_dict_form(bars_tuple_form):
        bars_dict_form = {}
        for dim, bar in bars_tuple_form:
            if dim not in bars_dict_form:
                bars_dict_form[dim] = []
            bars_dict_form[dim].append(bar)
        return bars_dict_form

    @staticmethod
    def process_input_to_dict_form(input_form):
        if len(input_form) == 0:
            return {}
        if isinstance(input_form, dict):
            return input_form
        list_form = list(input_form)
        if len(list_form[0]) == 2:
            if isinstance(list_form[0][0], int):
                try:
                    second_coordinate_length = len(list_form[0][1])
                except TypeError:
                    second_coordinate_length = -1
                if second_coordinate_length == 2:
                    return PlottingUtils.convert_tuple_form_to_dict_form(list_form)
            return {'' : list_form}
        raise TypeError("Bars given in an invalid format. Should be either a dim -> list_of_bars dictionary"
                        " or a (dim, (birth, death)) list.")

    @staticmethod
    def split_finite_infinite_single_diagram(bars_list):
        return ([bar for bar in bars_list if bar[1] < float('inf')],
                [bar for bar in bars_list if bar[1] == float('inf')])

    @staticmethod
    def split_finite_infinite_dictionary(bars_dict):
        bars_finite = {}
        bars_infinite = {}
        for dim in bars_dict:
            bars_finite[dim], bars_infinite[dim] = PlottingUtils.split_finite_infinite_single_diagram(bars_dict[dim])
        return bars_finite, bars_infinite

    @staticmethod
    def add_counts_text_to_markers(axs, rtol=1e-5, atol=1e-8):
        """Given axes, counts markers close to each other and adds the counts texts"""
        if isinstance(axs, Axes):
            axs = [axs]
        axs = np.array(axs).flatten()

        xlim, ylim = axs[0].get_xlim()[1] - axs[0].get_xlim()[0], axs[0].get_ylim()[1] - axs[0].get_ylim()[0]

        for ax in axs.flatten():
            for child in ax.get_children():
                if str(child.get_label()).isnumeric():
                    # extract points with frequencies
                    data = {}
                    for point in zip(*child.get_data()):
                        for key in data:
                            if (np.isclose(point[0], key[0], rtol=rtol, atol=atol)
                                    and np.isclose(point[1], key[1], rtol=rtol, atol=atol)):
                                data[key] += 1
                                break
                        else:
                            data[point] = 1
                    # print texts
                    for point, frequency in data.items():
                        if point[0] < .0 or point[1] > .9 * ylim:
                            posx = point[0] + .01 * xlim if point[0] < .9 * xlim else point[0] - .035 * xlim
                            posy = point[1] - .04 * ylim
                        else:
                            posx = point[0] + .005 * xlim
                            posy = point[1] + .015 * ylim
                        ax.text(posx, posy, frequency, fontdict={'color': child.get_color()})



def singleton(cls):
    __INSTANCES = {}

    def get_instance(*args, **kwargs):
        if cls not in __INSTANCES:
            __INSTANCES[cls] = cls(*args, **kwargs)
        return __INSTANCES[cls]

    return get_instance



import numpy as np

from chromatic_tda.utils.timing import TimingUtils


def intersect_lines(p, p_dir, q, q_dir):
    """Return intersection of two lines in 2D given by a point and
    a direction.
    Return None if the directions are parallel.
    """
    TimingUtils().start("GeometricalUtils - intersect lines")
    t,s = np.linalg.solve(np.transpose((p_dir, -q_dir)), (q-p)) # p + t*p_dir = q + s*q_dir
    TimingUtils().stop("GeometricalUtils - intersect lines")
    return p + t*p_dir


def project_to_line(p, p_dir, q):
    """Return orthogonal projection of point q into 2D line given
    by a point p and a direction p_dir."""
    TimingUtils().start("GeometricalUtils - project to line")
    res = intersect_lines(p, p_dir, q, np.array([-p_dir[1],p_dir[0]]))
    TimingUtils().stop("GeometricalUtils - project to line")
    return res


def bisector(p, q):
    """Return the bisector of points p,q as a point and a direction.
    The point is the midpoint between p and q."""
    TimingUtils().start("GeometricalUtils - bisector")
    dif = q-p
    res = (q+p)/2, np.array([-dif[1], dif[0]])
    TimingUtils().stop("GeometricalUtils - bisector")
    return res


def circum_center(p, q, r):
    """Return circum center of the triangle pqr."""
    TimingUtils().start("GeometricalUtils - circum center")
    res = intersect_lines(*bisector(p, q), *bisector(q, r))
    TimingUtils().stop("GeometricalUtils - circum center")
    return res


def sq_dist(p, q):
    """Return squared distance of two points."""
    TimingUtils().start("GeometricalUtils - sq dist")
    if len(p)!=len(q):
        raise ValueError("Both points need to be of the same dimension.")
    res = sum(i**2 for i in p-q)
    TimingUtils().stop("GeometricalUtils - sq dist")
    return res


def sq_dist_max(p, *args):
    """Return the max squared distance of p and points in args"""
    TimingUtils().start("GeometricalUtils - sq dist max")
    res = max((sq_dist(p, q) for q in args), default=0)
    TimingUtils().stop("GeometricalUtils - sq dist max")
    return res
    pass



import unittest
import numpy as np

from chromatic_tda.utils.linear_algebra_utils import LinAlgUtils


class LinearAlgebraTest(unittest.TestCase):
    a_mat = np.array([
        [7.56248052, 3.01108843, 0.56261590, 0.32744947, 0.23429895, 0.18101955, 0.70902646, 0.10934863, 0.9035789],
        [4.34061090, 1.47696429, 0.91088442, 0.99761031, 0.65215561, 0.42838886, 0.62356712, 0.97818994, 0.9252501],
        [1.58343582, 0.62058533, 0.24465723, 0.92581940, 0.98481243, 0.94776809, 0.39252661, 0.84119163, 0.8511891],
        [2.80497347, 2.49306469, 0.97397928, 0.55287025, 0.77001101, 0.59266880, 0.71048821, 0.43799465, 0.0107083]
    ])
    b_vec = np.array([0.57413191, 0.79050492, 2.42991792, 8.25227247])

    def test_solution_full_rank_4x9(self):
        a_mat = LinearAlgebraTest.a_mat
        b_vec = LinearAlgebraTest.b_vec
        x, ker = LinAlgUtils.solve(a_mat, b_vec)

        assert np.isclose(a_mat @ x, b_vec).all()
        assert len(ker) == 5
        for v in ker:
            assert np.isclose(a_mat @ (x + v), b_vec).all()

    def test_solution_singular_8x9_stack(self):
        a_mat = np.concatenate([LinearAlgebraTest.a_mat, LinearAlgebraTest.a_mat])
        b_vec = np.concatenate([LinearAlgebraTest.b_vec, LinearAlgebraTest.b_vec])
        x, ker = LinAlgUtils.solve(a_mat, b_vec)

        assert np.isclose(a_mat @ x, b_vec).all()
        assert len(ker) == 5
        for v in ker:
            assert np.isclose(a_mat @ (x + v), b_vec).all()

    def test_solution_singular_7x9(self):
        coef_mat = np.array([
            [1, 2, 3, 4],
            [4, 3, 2, 1],
            [0.3005, 0.6383, 0.6640, 0.9111]
        ])
        a_mat = np.concatenate([LinearAlgebraTest.a_mat, coef_mat @ LinearAlgebraTest.a_mat])
        b_vec = np.concatenate([LinearAlgebraTest.b_vec, coef_mat @ LinearAlgebraTest.b_vec])
        x, ker = LinAlgUtils.solve(a_mat, b_vec)

        assert np.isclose(a_mat @ x, b_vec).all()
        assert len(ker) == 5
        for v in ker:
            assert np.isclose(a_mat @ (x + v), b_vec).all()

    def test_solution_singular_4x12(self):
        coef_mat = np.array([
            [0.0262, 0.8473, 0.5723, 0.4596, 0.2175, 0.7477, 0.4135, 0.7733, 0.8522],
            [0.0933, 0.9487, 0.2949, 0.5585, 0.1369, 0.3590, 0.7447, 0.9125, 0.3629],
            [0.4308, 0.2242, 2.2360, 0.7456, 0.6092, 0.7564, 5.9848, 0.6998, 0.8237]
        ]).transpose()
        a_mat = np.concatenate([LinearAlgebraTest.a_mat, LinearAlgebraTest.a_mat @ coef_mat], axis=1)
        b_vec = LinearAlgebraTest.b_vec
        x, ker = LinAlgUtils.solve(a_mat, b_vec)

        assert np.isclose(a_mat @ x, b_vec).all()
        assert len(ker) == 8
        for v in ker:
            assert np.isclose(a_mat @ (x + v), b_vec).all()

    def test_solution_full_rank_9x4(self):
        a_mat = LinearAlgebraTest.a_mat.transpose()
        b_vec = a_mat @ [1, 2, 3, 4]

        x, ker = LinAlgUtils.solve(a_mat, b_vec)

        assert np.isclose(x, [1, 2, 3, 4]).all()
        assert len(ker) == 0

    def test_solution_singular_9x7(self):
        coef_mat = np.array([
            [1, 2, 3, 4],
            [4, 3, 2, 1],
            [0.3005, 0.6383, 0.6640, 0.9111]
        ])
        a_mat = np.concatenate([LinearAlgebraTest.a_mat, coef_mat @ LinearAlgebraTest.a_mat]).transpose()
        b_vec = a_mat @ [1, 2, 3, 4, 0, 0, 0]

        x, ker = LinAlgUtils.solve(a_mat, b_vec)

        assert np.isclose(a_mat @ x, b_vec).all()
        assert len(ker) == 3
        for v in ker:
            assert np.isclose(a_mat @ (x + v), b_vec).all()

    def test_orthogonalize_4x9(self):
        a_mat = self.a_mat
        ortho = LinAlgUtils.orthogonalize_rows(a_mat)
        rank_ortho = (~np.isclose(np.linalg.svd(ortho).S, 0)).sum()
        rank_ortho_a_mat = (~np.isclose(np.linalg.svd(np.concatenate([a_mat, ortho])).S, 0)).sum()

        assert ortho.shape == (4, 9)
        assert rank_ortho == 4 and rank_ortho_a_mat == 4  # row spaces match
        assert np.isclose(ortho @ ortho.transpose(), np.identity(4)).all()  # rows are orthonormal

    def test_orthogonalize_2x9(self):
        a_mat = self.a_mat[:2]
        ortho = LinAlgUtils.orthogonalize_rows(a_mat)
        rank_ortho = (~np.isclose(np.linalg.svd(ortho).S, 0)).sum()
        rank_ortho_a_mat = (~np.isclose(np.linalg.svd(np.concatenate([a_mat, ortho])).S, 0)).sum()

        assert ortho.shape == (2, 9)
        assert rank_ortho == 2 and rank_ortho_a_mat == 2  # row spaces match
        assert np.isclose(ortho @ ortho.transpose(), np.identity(2)).all()  # rows are orthonormal

    def test_orthogonalize_singular_7x9(self):
        coef_mat = np.array([
            [1, 2, 3, 4],
            [4, 3, 2, 1],
            [0.3005, 0.6383, 0.6640, 0.9111]
        ])
        a_mat = np.concatenate([LinearAlgebraTest.a_mat, coef_mat @ LinearAlgebraTest.a_mat])
        ortho = LinAlgUtils.orthogonalize_rows(a_mat)
        rank_ortho = (~np.isclose(np.linalg.svd(ortho).S, 0)).sum()
        rank_ortho_a_mat = (~np.isclose(np.linalg.svd(np.concatenate([a_mat, ortho])).S, 0)).sum()

        assert ortho.shape == (4, 9)
        assert rank_ortho == 4 and rank_ortho_a_mat == 4  # row spaces match
        assert np.isclose(ortho @ ortho.transpose(), np.identity(4)).all()  # rows are orthonormal



from .entities.chromatic_alpha_complex import ChromaticAlphaComplex
from .entities.simplicial_complex import SimplicialComplex
from .plots.plotting_functions import plot_persistence_diagram, plot_six_pack, plot_labeled_point_set, plot_edges
from .experimental.feature_extraction import BLANK
import importlib.metadata

__version__ = importlib.metadata.version("chromatic_tda")



import numpy as np
import numpy.typing as npt
from collections.abc import KeysView, ValuesView

from chromatic_tda.utils.boundary_matrix_utils import BoundaryMatrixUtils
from chromatic_tda.utils.simplex_utils import SimplexUtils
from chromatic_tda.utils.singleton import singleton

from itertools import combinations
from chromatic_tda.core.core_simplicial_complex import CoreSimplicialComplex


@singleton
class CoreSimplicialComplexFactory:
    EMPTY_COMPLEX_DIMENSION = -1

    def create_instance(self, simplices) -> CoreSimplicialComplex:
        simplicial_complex = CoreSimplicialComplex()
        self.build_complex(simplicial_complex, simplices)

        return simplicial_complex

    def build_complex(self, simplicial_complex: CoreSimplicialComplex, simplices):
        if isinstance(simplices, dict):
            self.build_complex_from_dictionary(simplicial_complex, simplices)
        elif isinstance(simplices, (list, tuple, set, np.ndarray, KeysView, ValuesView)):
            self.build_complex_from_list(simplicial_complex, simplices)
        else:
            raise TypeError(f"Cannot build complex from type {type(simplices)}.")

    def build_complex_from_dictionary(self, simplicial_complex: CoreSimplicialComplex, simplex_weights_dict):
        self.build_complex_from_list(simplicial_complex, simplex_weights_dict.keys())
        simplicial_complex.set_simplex_weights(simplex_weights_dict)

    def build_complex_from_list(self, simplicial_complex: CoreSimplicialComplex, simplex_list):
        simplicial_complex.dimension = self.find_maximal_dimension(simplex_list)
        simplicial_complex.dim_simplex_dict = self.build_dimension_dictionary(
            simplex_list, max_dimension=simplicial_complex.dimension)
        self.add_boundary_and_missing_simplices(simplicial_complex)
        simplicial_complex.set_simplex_weights({})

    @staticmethod
    def find_maximal_dimension(simplex_list):
        return max((SimplexUtils.dimension(simplex) for simplex in simplex_list),
                   default = CoreSimplicialComplexFactory().EMPTY_COMPLEX_DIMENSION)

    @staticmethod
    def build_dimension_dictionary(simplex_list, max_dimension):
        dim_simplex_dict = {dim : set() for dim in range(max_dimension, -1, -1)}
        for simplex in simplex_list:
            simplex_dim = SimplexUtils.dimension(simplex)
            dim_simplex_dict[simplex_dim].add(tuple(sorted(simplex)))
        return dim_simplex_dict

    @staticmethod
    def add_boundary_and_missing_simplices(simplicial_complex: CoreSimplicialComplex):
        for dim in range(simplicial_complex.dimension, 0, -1):
            for simplex in simplicial_complex.dim_simplex_dict[dim]:
                simplicial_complex.boundary[simplex] = set(combinations(simplex, dim))
            simplicial_complex.dim_simplex_dict[dim - 1] = simplicial_complex.dim_simplex_dict[dim - 1].union(
                *[simplicial_complex.boundary[simplex] for simplex in simplicial_complex.dim_simplex_dict[dim]])
        for vertex in simplicial_complex.dim_simplex_dict[0]:
            simplicial_complex.boundary[vertex] = set()

    @staticmethod
    def create_restricted_instance(simplicial_complex: CoreSimplicialComplex,
                                   restricted_simplices) -> CoreSimplicialComplex:
        """Return a new SimplicialComplex restricted to given simplices."""
        new_complex : CoreSimplicialComplex = CoreSimplicialComplex()
        restricted_simplices_set : set = set(restricted_simplices)

        new_complex.dim_simplex_dict = {d : simplicial_complex.dim_simplex_dict[d] & restricted_simplices_set
                                        for d in simplicial_complex.dim_simplex_dict}
        new_complex.clear_empty_dimensions()
        new_complex.dimension = max(new_complex.dim_simplex_dict,
                                    default = CoreSimplicialComplexFactory().EMPTY_COMPLEX_DIMENSION)

        new_complex.boundary = {simplex: simplicial_complex.boundary[simplex] & restricted_simplices_set
                                for simplex in set(simplicial_complex.boundary) & restricted_simplices_set}
        new_complex.co_boundary = BoundaryMatrixUtils.make_co_boundary(new_complex.boundary)
        new_complex.simplex_weights = {simplex : simplicial_complex.simplex_weights[simplex]
                                       for simplex in new_complex.boundary}
        new_complex.sub_complex = simplicial_complex.sub_complex & restricted_simplices_set

        return new_complex



class FilterFunctions:
    """Functions to order tuples according to float values and subset-relation."""
    def __init__(self, weight_function : dict, sub_complex : set):
        self.weight_function = weight_function
        self.sub_complex = sub_complex
        self.total_filtration = {simplex : index for index, simplex in enumerate(
            sorted(weight_function.keys(), key= lambda simplex: (weight_function[simplex], len(simplex), simplex))
        )}

    def filter_function_rad(self):
        """Filter by radius."""
        return lambda simplex: self.total_filtration[simplex]

    def filter_function_rad_sub_first(self):
        """Filter by radius, but put all sub-complex simplices first, and then the rest."""
        return lambda simplex: (simplex not in self.sub_complex, self.total_filtration[simplex])



import numpy as np
import numpy.typing as npt
from numpy.linalg._linalg import SVDResult

from chromatic_tda.utils.floating_point_utils import FloatingPointUtils
from chromatic_tda.utils.timing import TimingUtils


class LinAlgUtils:
    @staticmethod
    def solve(a_matrix: npt.NDArray, b_vector: npt.NDArray, check_solution=False) -> tuple[npt.NDArray, npt.NDArray]:
        """
        Solve a general linear equation Ax=b using singular value decomposition.
        Particular solution computed via pseudo-inverse A^+ as A^+ @ b. This makes sense even if the equation
        has no solution, and returns the least squares solution. By default, no warning is given -- see check_solution.

        :param a_matrix:        matrix A
        :param b_vector:        right side b
        :param check_solution:  if True, Ax == b check is performed for the particular solution x,
                                and error raised if it fails

        :return: A tuple (x, kernel) where x is a particular solution, and kernel is an orthonormal basis of Ker(A)
        (vectors in rows).
        """
        TimingUtils().start("LinAlg :: Solve Linear Equation With Kernel")
        svd : SVDResult = np.linalg.svd(a_matrix)  # svd.U @ diagonal from svd.S @ svd.Vh == A
        rank : int = LinAlgUtils.count_nonzero(svd.S)

        pseudoinverse = LinAlgUtils.pseudoinverse_from_svd(svd)
        x = pseudoinverse @ b_vector
        kernel = svd.Vh[rank :]  # the vectors that Vh sends to E_{rank+1}, ..., E_{n}: Vhh @ E_i is i-th row of Vh

        if check_solution and not LinAlgUtils.check_solution(a_matrix, b_vector, x):
            raise np.linalg.LinAlgError("There is no exact solution to given Ax=b.")

        TimingUtils().stop("LinAlg :: Solve Linear Equation With Kernel")
        return x, kernel

    @staticmethod
    def orthogonalize_rows(array: npt.NDArray) -> npt.NDArray:
        """Given m x n array A with m <= n,
        return array whose rows are orthonormal basis of the row-space of A.
        WARNING: for rank(A) < m might return smaller space if R in np.linalg.qr would skip a pivot and use it later"""
        TimingUtils().start("LinAlg :: Orthogonalize Rows")
        m, n = array.shape
        if m > n:
            raise ValueError("Only m x n arrays with m <= n allowed")
        qr = np.linalg.qr(array.transpose(), mode='reduced')
        non_zero_r_diagonal = ~np.array([FloatingPointUtils.is_close(x, 0) for x in qr.R.diagonal()], dtype=bool)
        TimingUtils().stop("LinAlg :: Orthogonalize Rows")
        return qr.Q.transpose()[non_zero_r_diagonal]

    @staticmethod
    def count_nonzero(array: npt.NDArray) -> int:
        return int(np.prod(array.shape)) - sum(int(FloatingPointUtils.is_close(0, x)) for x in array)

    @staticmethod
    def check_solution(a_matrix: npt.NDArray, b_vector: npt.NDArray, x_vector: npt.NDArray):
        return FloatingPointUtils.is_all_close(a_matrix @ x_vector, b_vector)

    @staticmethod
    def pseudoinverse_from_svd(svd: SVDResult) -> npt.NDArray:
        TimingUtils().start("LinAlg :: Pseudoinverse From SVD")
        pseudo_s = np.zeros(shape=(svd.U.shape[1], svd.Vh.shape[0]))
        np.fill_diagonal(pseudo_s, LinAlgUtils.pseudoinverse_of_diagonal(svd.S))
        pseudoinverse = (svd.U @ pseudo_s @ svd.Vh).transpose()
        TimingUtils().stop("LinAlg :: Pseudoinverse From SVD")
        return pseudoinverse

    @staticmethod
    def pseudoinverse_of_diagonal(diagonal: npt.NDArray) -> npt.NDArray:
        return np.array([s ** -1 if not FloatingPointUtils.is_close(s, 0) else 0 for s in diagonal])



import multiprocessing
import math
import os

from multiprocessing.managers import SharedMemoryManager
from chromatic_tda.utils.timing import TimingUtils

N = 5*1000*1000

class ParallelTestCore():
    def __init__(self):
        self.atoms = os.cpu_count()
        self.actions = math.ceil(N/self.atoms)

    def cube(self, x):
        return math.sqrt(x)*math.sqrt(x)*math.sqrt(x)*math.sqrt(x)*math.sqrt(x)*math.sqrt(x)*math.sqrt(x)

class LinearTest(ParallelTestCore):
    def run(self):
        TimingUtils().start("Linear")
        result = {}
        for x in range(0, N):
            result[x] = self.cube(x)
        print(len(result))
        print(result[int(N/2)])
        TimingUtils().stop("Linear")

class ParallelWithPoolTest(ParallelTestCore):

    def cube_atoms(self, x):
        print(f"atom {x}")
        res = []
        for i in range(0, self.actions):
            k = x * self.actions + i
            if k >= N:
                break
            res.append(self.cube(k))
        return res        

    def run(self):
        TimingUtils().start("Pool")
        result = []
        with multiprocessing.Pool() as pool:
            res_list = pool.map(self.cube_atoms, range(0, self.atoms))
            for res in res_list:
                result.extend(res)
        print(len(result))
        #print(result[int(N/2)])
        TimingUtils().stop("Pool")


class ParallelWithProcessTest(ParallelTestCore):

    def cube_atoms(self, x, res):
        result = {}
        for i in range(0, self.actions):
            k = x * self.actions + i
            if k >= N:
                break
            result[k] = self.cube(k)
        res[x] = result

    def run(self):
        TimingUtils().start("Process")
        with SharedMemoryManager() as smm:
            process_dict = {}
            result = {}
            process_res = smm.ShareableList(range(self.atoms))
            for x in range(0, self.atoms):
                process = multiprocessing.Process(target= self.cube_atoms, args=[x, process_res])
                process.start()            
                process_dict[x] = process

            for x in range(0, self.atoms):
                process = process_dict[x]
                process.join()
        
                result = result | process_res[x]
    
        print(len(result))
        TimingUtils().stop("Process")


def parallel():
    LinearTest().run()
    ParallelWithPoolTest().run()
#    ParallelWithProcessTest().run()
    TimingUtils().print()

if __name__ == "__main__":
    parallel()


import numpy as np
from dataclasses import dataclass

from chromatic_tda.utils.floating_point_utils import FloatingPointUtils


@dataclass
class StackOfSpheres:
    center: np.array
    radii: dict[any, float]
    maximum_radius: float
    maximum_labels: set

    def __init__(self, center: np.array, radii: dict[any, float]):
        self.center = center
        self.radii = radii
        self.maximum_radius = max(radii.values())
        self.maximum_labels = {label for label, radius in self.radii.items()
                               if FloatingPointUtils.is_close(radius, self.maximum_radius)}



import unittest
import numpy as np

from chromatic_tda.utils.geometry_utils import GeometryUtils, AffineSpace


class GeometryTestReflections(unittest.TestCase):
    def test_reflection_3d_through_3d(self):  # identity
        # Reflections also test orthogonal projection
        vector_space = np.array([
            [1, 1, 0],
            [2, 1, 0],
            [.4, -.123, .03]
        ])
        shift = np.array([1, 1, 1])
        points = [
            np.array([2, 3, 4]),
            np.array([0, 0, 0]),
            np.array([1, 1, 1]),
            np.array([5, 6, 1]),
            np.array([2, 1, -3]),
            np.array([1.2, 3.1, 2.789])
        ]
        reflected_points = GeometryUtils.reflect_points_through_affine_space(AffineSpace(shift, vector_space), *points)
        for pt, pt_ref in zip(points, reflected_points):
            assert np.isclose(pt, pt_ref).all()

    def test_reflection_3d_through_2d(self):
        vector_space = np.array([
            [1, 1, 0],
            [2, 1, 0]
        ])
        shift = np.array([1, 1, 1])
        points = [
            np.array([2, 3, 4]),
            np.array([0, 0, 0]),
            np.array([1, 1, 1]),
            np.array([5, 6, 1]),
            np.array([2, 1, -3]),
            np.array([1.2, 3.1, 2.789])
        ]
        reflected_points = GeometryUtils.reflect_points_through_affine_space(AffineSpace(shift, vector_space), *points)
        for pt, pt_ref in zip(points, reflected_points):
            assert np.isclose((pt - (0, 0, 1)) * (1, 1, -1) + (0, 0, 1), pt_ref).all()

    def test_reflection_3d_through_2d_rotated(self):
        rot_angle = .7
        rotation = np.array([
            [1, 0, 0],
            [0, np.cos(rot_angle), - np.sin(rot_angle)],
            [0, np.sin(rot_angle), np.cos(rot_angle)]
        ])
        vector_space = np.array([
            rotation @ [1, 1, 0],
            rotation @ [2, 1, 0]
        ])
        shift = np.array([1, 1, 1])
        points = [
            np.array([2, 3, 4]),
            np.array([0, 0, 0]),
            np.array([1, 1, 1]),
            np.array([5, 6, 1]),
            np.array([2, 1, -3]),
            np.array([1.2, 3.1, 2.789])
        ]
        reflected_points = GeometryUtils.reflect_points_through_affine_space(AffineSpace(shift, vector_space), *points)
        for pt, pt_ref in zip(points, reflected_points):
            assert np.isclose(rotation @ ((rotation.transpose() @ (pt - (1, 1, 1))) * (1, 1, -1)) + (1, 1, 1),
                              pt_ref).all()

    def test_reflection_3d_through_2d_single_point(self):
        vector_space = np.array([
            [1, 1, 1],
            [0, 0, 1]
        ])
        shift = np.array([1, 1, 1])
        point = np.array([1, 0, 0])
        reflected_points = GeometryUtils.reflect_points_through_affine_space(AffineSpace(shift, vector_space), point)
        assert len(reflected_points) == 1
        assert np.isclose(reflected_points[0], (0, 1, 0)).all()

    def test_reflection_3d_through_1d(self):
        vector_space = np.array([
            [0, 0, 102.3],
        ])
        shift = np.array([5, 6, 1])
        points = [
            np.array([2, 3, 4]),
            np.array([0, 0, 0]),
            np.array([1, 1, 1]),
            np.array([5, 6, 1]),
            np.array([2, 1, -3]),
            np.array([1.2, 3.1, 2.789])
        ]
        reflected_points = GeometryUtils.reflect_points_through_affine_space(AffineSpace(shift, vector_space), *points)
        for pt, pt_ref in zip(points, reflected_points):
            assert np.isclose((pt - (5, 6, 0)) * (-1, -1, 1) + (5, 6, 0), pt_ref).all()

    def test_reflection_3d_through_0d(self):  # reflection through the point [1, 1, 1]
        vector_space = np.zeros(shape=(0, 3))
        shift = np.array([1, 1, 1])
        points = [
            np.array([2, 3, 4]),
            np.array([0, 0, 0]),
            np.array([1, 1, 1]),
            np.array([5, 6, 1]),
            np.array([2, 1, -3]),
            np.array([1.2, 3.1, 2.789])
        ]
        reflected_points = GeometryUtils.reflect_points_through_affine_space(AffineSpace(shift, vector_space), *points)
        for pt, pt_ref in zip(points, reflected_points):
            assert np.isclose(- (pt - (1, 1, 1)) + (1, 1, 1), pt_ref).all()

    def test_circumcircle_of_weighted_basic_triangle(self):
        points = np.array([
            [1, 2, 3],
            [1, 1, 1],
            [4, 3, 2]
        ])
        weights = np.array([0, 0, 0])
        x, rad2 = GeometryUtils.circumsphere_of_weighted_points(points, weights)
        assert np.allclose(np.square(points - x).sum(axis=1), rad2)

    def test_circumcircle_of_weighted_basic_weighted_triangle(self):
        points = np.array([
            [1, 2, 3],
            [1, 1, 1],
            [4, 3, 2]
        ])
        weights = np.array([.3, -1.1, 0])
        x, rad2 = GeometryUtils.circumsphere_of_weighted_points(points, weights)
        print(np.square(points[0] - x).sum() + weights[0])
        print(np.square(points - x).sum(axis=1) + weights)
        print(rad2)
        assert np.allclose(np.square(points - x).sum(axis=1) + weights, rad2)

    def test_circumcircle_of_weighted_4d_line(self):
        points = np.array([
            [1, 2, 3, 0],
            [1, 0, 1, 1]
        ])
        weights = np.array([0, 0])
        x, rad2 = GeometryUtils.circumsphere_of_weighted_points(points, weights)
        assert np.allclose(np.square(points - x).sum(axis=1), rad2)
        assert np.isclose(9 / 4, rad2)

    def test_circumcircle_of_weighted_4d_weighted_line(self):
        points = np.array([
            [1, 2, 3, 0],
            [1, 0, 1, 1]
        ])
        weights = np.array([2, 5])
        x, rad2 = GeometryUtils.circumsphere_of_weighted_points(points, weights)
        assert np.allclose(np.square(points - x).sum(axis=1) + weights, rad2)



from chromatic_tda.core.core_chromatic_alpha_complex import CoreChromaticAlphaComplex
from chromatic_tda.core.chromatic_alpha_complex_factory import CoreChromaticAlphaComplexFactory, \
    CoreChromaticAlphaComplexTorus2DFactory
from chromatic_tda.entities.simplicial_complex import SimplicialComplex


class ChromaticAlphaComplex:

    def __init__(self, points, labels, lift_perturbation=1e-9, point_perturbation=None, **kwargs) -> None:
        """Create an instance of ChromaticAlphaComplex. The object contains the full chromatic Delaunay complex together
        with its alpha radius function.

        Arguments:
            points ... List of point coordinates
            labels ... List of labels. The length has to be the same as the number of points.
                       The first label goes with the first point, etc... The labels can be any hashable object.

        Keyword arguments:
            point_perturbation ... Perturb the points given on the input by +-point_perturbation/2 in each coordinate.
                                   (default: None)
            lift_perturbation ... Compute the Delaunay complex with perturbed lifting to break the non-general position.
                                  Generally leads to faster computation, as QHull does not need to deal with the
                                  non-generality itself. (default: 1e-9)
        """
        if kwargs.get('torus', False):
            factory = CoreChromaticAlphaComplexTorus2DFactory(points, labels,
                                                              xrange= kwargs.get('xrange', None),
                                                              yrange= kwargs.get('yrange', None),
                                                              suppress_wrapping_check= kwargs.get(
                                                                  'suppress_wrapping_check', False),
                                                              suppress_boundary_consistency_check= kwargs.get(
                                                                  'suppress_boundary_consistency_check', False))
        else:
            factory = CoreChromaticAlphaComplexFactory(points, labels)
        self.core_alpha_complex : CoreChromaticAlphaComplex = factory.create_instance(
            lift_perturbation=lift_perturbation, point_perturbation=point_perturbation)

    def __iter__(self):
        yield from self.core_alpha_complex

    def __len__(self) -> int:
        return len(self.core_alpha_complex)

    def __contains__(self, element) -> bool:
        return element in self.core_alpha_complex

    def get_simplicial_complex(self, sub_complex=None, full_complex=None, relative=None,
                               allow_unused_labels=False) -> SimplicialComplex:
        """Generate a simplicial complex and sub-complex pair based on the parameters given.
        The parameter complex restricts the complex, the parameter sub_complex defines the sub-complex,
        and the parameter relative erases simplices from the complex to represent a relative simplicial complex.

        Each parameter can be given as a list of lists of labels. Each simplex is kept (for sub_complex and complex)
        or erased (for relative) iff the set of labels of its vertices is contained in one of the lists of labels.
        For example sub_complex=[['blue', 'red'], ['blue', 'green']] chooses those simplices whose vertices are labeled
        either 'blue' and 'red' or 'blue' and 'green', but NOT 'red' and 'green'; also NOT all three colors,
        'blue','red','green'. The single-color simplices, e.g. all 'blue', ARE chosen.

        If the labels are single-digit integers, the parameters can also be given as lists of words separated by commas.
        For example `sub-complex = 01,02` will return complex with the sub-complex being all simplices with
        colors 0, 1, 01 or 02 -- in particular, no simplices with colors 12 or 012.

        Finally, the parameter can also be one of the following words:
            'all' ... allow all simplices
            'mono-chromatic' ... allow simplices containing only one label (='0,1,2' if labels are 0,1,2)
            'bi-chromatic' ... allow simplices containing at most two labels (='01,02,12' if labels are 0,1,2)
            'tri-chromatic' ... allow simplices containing at most three labels (='012'='all' if labels are 0,1,2)
            Can also be written as 'one-chromatic' or '1-chromatic' etc... The dash is optional.

        Keyword arguments:
        allow_unused_labels ... By default (False), an error is raised if the sub_complex/complex/relative parameters
                                contain a label that is not used. To suppress this behavior, set this parameter to True.
                                In that case the unused labels make no difference on the result.
        """
        return SimplicialComplex(self.core_alpha_complex.get_simplicial_complex(
            sub_complex=sub_complex, full_complex=full_complex, relative=relative,
            allow_unused_labels=allow_unused_labels))

    def weight_function(self, simplex=None):
        """If simplex is given, return the weight/radius of the simplex.
        If no simplex is given, return the weight/radius function as a dictionary {simplex : weight}."""
        if simplex is None:
            return self.core_alpha_complex.simplicial_complex.get_weight_function_copy()
        return self.core_alpha_complex.simplicial_complex.get_simplex_weight(simplex)

    def simplices(self):
        """Return list of all simplices sorted by dimension and then lexicographically (w.r.t. vertex indices)."""
        return set(self.core_alpha_complex.simplicial_complex.get_simplices())

    def simplex_labels(self, simplex):
        """Return set of labels of the vertices of the given simplex"""
        if simplex in self:
            return self.core_alpha_complex.simplex_labels_input(simplex)
        else:
            raise KeyError(f'{simplex} is not a simplex in the chromatic alpha complex.')

    def simplex_points(self, simplex):
        """Return a list of points spanning the given simplex"""
        if simplex in self:
            return [self.core_alpha_complex.points[vertex] for vertex in sorted(simplex)]
        else:
            raise KeyError(f'{simplex} is not a simplex in the chromatic alpha complex.')

    def points(self):
        """Return the point cloud defining the complex."""
        return self.core_alpha_complex.copy_points()

    def labels(self):
        """Return the labels for the points defining the complex."""
        return [self.core_alpha_complex.internal_labels_to_input_labels_dict[lab]
                for lab in self.core_alpha_complex.internal_labeling]



from pathlib import Path
import unittest
import json
import numpy as np

from chromatic_tda import ChromaticAlphaComplex
from chromatic_tda.utils.floating_point_utils import FloatingPointUtils


# class TestClass(unittest.TestCase):
#
#     def test_something(self):
#         TestBars().test_all()


class TestBars(unittest.TestCase):

    data_folder = Path(__file__).parent / 'test_data'

    def test_precomputed_line_sep(self, verbose=False, assertions=True):
        self.single_test('line_sep', verbose, assertions)

    def test_precomputed_one_circle_20_40(self, verbose=False, assertions=True):
        self.single_test('one_circle_20_40', verbose, assertions)

    def test_precomputed_two_circles(self, verbose=False, assertions=True):
        self.single_test('two_circles', verbose, assertions)

    def test_precomputed_two_circles_cc(self, verbose=False, assertions=True):
        self.single_test('two_circles_cc', verbose, assertions)

    def test_precomputed_random(self, verbose=False, assertions=True):
        self.single_test('random', verbose, assertions)

    def test_precomputed_random2(self, verbose=False, assertions=True):
        self.single_test('random2', verbose, assertions)

    # def test_precomputed_one_circle_3col_back_split(self, verbose=False, assertions=True):
    #     self.single_test('one_circle_3col_back_split', verbose, assertions)
    #
    # def test_precomputed_one_circle_3col_circ_split(self, verbose=False, assertions=True):
    #     self.single_test('one_circle_3col_circ_split', verbose, assertions)

    def test_precomputed_one_circle_3col_bi_circle_tri_filled(self, verbose=False, assertions=True):
        self.single_test('one_circle_3col_bi-circle_tri-filled', verbose, assertions)

    def test_custom_colors(self, verbose=False, assertions=True):
        # an instance with labels 'blue', 'orange' instead of 0, 1
        data = self.load_data('one_circle_20_40')
        data['labels'] = ['blue' if lab == 0 else 'orange' for lab in data['labels']]
        for instance in data['persistence']:
            instance['map']['sub_complex'] = [
                [{'0': 'blue', '1': 'orange'}[ch]] for ch in instance['map']['sub_complex'].split(',')]
        test = self.single_test_data(data, return_detailed=False)
        if assertions:
            assert test
        if verbose:
            print(f'    {test}  ...  one_circle_20_40 recolored')

    def test_close_stack_radii_artifact_2hom(self):
        points = (
            [0.09459416951153077, 0.11232276848750489],  # 0
            [0.9680112771867585, 0.15811534146081496],   # 1
            [0.9658547896521128, 0.1702045941509115],    # 2
            [0.8312718291955721, 0.6406528236300497],    # 3
            [0.09659691184397712, 0.11313859402162363],  # 4
            [0.9595099529549772, 0.15846022651540637],   # 5
            [0.8425246776313106, 0.635219429110198],     # 6
            [0.9701160920689117, 0.16062702956075603],   # 7
            [0.8384241285861569, 0.6479053787601031],    # 8
            [0.10089735680674439, 0.12333065671267851],  # 9
            [0.831030327917524, 0.6343887895213592],     # 10
            [0.08917225973887621, 0.1294156525294371]    # 11
        )
        labels = (0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1, 1)
        cplx = ChromaticAlphaComplex(points, labels).get_simplicial_complex(sub_complex='0')
        cplx.compute_persistence()
        cplx_2_bars = [(b, d) for b, d in cplx.core_complex.birth_death['complex']['pairs']
                       if cplx.weight_function(b) < cplx.weight_function(d)
                       and len(b) == 3]
        assert len(cplx_2_bars) == 0

    def single_test(self, data_name, verbose=False, assertions=True):
        test = self.single_test_data(self.load_data(data_name), return_detailed=False)
        if assertions:
            assert test
        if verbose:
            print(f'    {test}  ...  {data_name}')

    def load_data(self, data_name):
        with open(self.data_folder / f'{data_name}.json', 'r') as file:
            data = json.load(file)

        return data

    @staticmethod
    def single_test_data(data, return_detailed=False):
        results = []
        alpha_complex = ChromaticAlphaComplex(data['points'], data['labels'])
        for instance in data['persistence']:
            results_instance = {}
            simplicial_complex = alpha_complex.get_simplicial_complex(
                full_complex = instance['map']['complex'],
                sub_complex = instance['map']['sub_complex'],
                relative = instance['map']['relative']
            )
            simplicial_complex.compute_persistence()
            for group in ('kernel', 'sub_complex', 'image', 'complex', 'cokernel', 'relative'):
                bars = simplicial_complex.bars(group, return_as='list')
                bars_test = [(dim, bar) for dim, bar in instance['bars'][group]
                             if not FloatingPointUtils.is_trivial_bar(bar)]
                results_instance[group] = all(
                    dim == dim_test and np.isclose(birth, birth_test) and np.isclose(death, death_test)
                    for (dim, (birth, death)), (dim_test, (birth_test, death_test)) in
                    zip(sorted(bars), sorted(bars_test))
                )
            results.append(results_instance)

        if return_detailed:
            return results
        else:
            return all(all(group_result for group_result in instance_results.values()) for instance_results in results)


def main():
    pass


if __name__ == "__main__":
    main()



import numpy as np

from chromatic_tda.utils.boundary_matrix_utils import BoundaryMatrixUtils
from chromatic_tda.utils.floating_point_utils import FloatingPointUtils


class CoreSimplicialComplex:
    dim_simplex_dict: dict[int, set]
    simplex_weights: dict[tuple, float]
    boundary: dict[tuple, set]
    co_boundary: dict[tuple, set]
    sub_complex: set
    persistence_data: dict
    birth_death: dict
    dimension: int

    def __init__(self) -> None:
        self.clear()

    def __iter__(self):
        yield from self.get_simplices()

    def __len__(self) -> int:
        return len(self.boundary)

    def __contains__(self, item) -> bool:
        return item in self.boundary

    def clear(self) -> None:
        self.dim_simplex_dict = {}
        self.simplex_weights = {}  # radius values of simplices

        self.boundary = {}  # example: {(1,2,3) : {(1,2), (1,3), (2,3)}, (1,2) : {(1), (2)}, ...}
        self.co_boundary = {}  # example:  {(1,2) : {(1,2,3), (1,2,4)}, (1,3) : {(1,2,3), (1,3,4)}, ...}

        self.sub_complex = set()
        self.persistence_data = {}
        self.birth_death = {}

        self.dimension = 0

    def clear_empty_dimensions(self) -> None:
        clear_dims = []
        for dim in self.dim_simplex_dict:
            if len(self.dim_simplex_dict[dim]) == 0:
                clear_dims.append(dim)
        for dim in clear_dims:
            self.dim_simplex_dict.pop(dim)

    def get_bars_list(self, group: str, dim: int = None, only_finite: bool = False) -> list:
        """
        Return list of pairs representing bars of a given dimension.
        If `dim` is not given, returns all bars in form `(dim, (birth, death))`.
        The groups to choose from (`group`) are:
            complex, sub_complex, image, kernel, cokernel, relative
        """
        if self.birth_death is None:
            raise ValueError("Persistence not yet computed, run `compute_persistence` first.")
        if group not in self.birth_death:
            raise ValueError(f"Persistence for `{group}` not computed. Did you run `compute_persistence`?")

        dim_bars_finite: list[tuple[int, tuple[float, float]]] = [
            (len(s) - 1 if group != 'kernel' else len(s) - 2,
             (self.simplex_weights[s], self.simplex_weights[t])) for s, t in self.birth_death[group]['pairs']
        ]
        if only_finite:
            dim_bars_infinite = []
        else:
            dim_bars_infinite: list[tuple[int, tuple[float, float]]] = [
                (len(s) - 1 if group != 'kernel' else len(s) - 2,
                 (self.simplex_weights[s], np.inf)) for s in self.birth_death[group]['essential']
            ]
        dim_bars = dim_bars_finite + dim_bars_infinite

        if dim is None:
            return sorted([b for b in dim_bars if not FloatingPointUtils.is_trivial_bar(b[1])])
        else:
            bars: list[tuple] = [bar for bar_dim, bar in dim_bars if bar_dim == dim]
            return sorted([b for b in bars if not FloatingPointUtils.is_trivial_bar(b)])

    def get_simplices(self) -> list:
        """Return list of all simplices sorted by dimension and then lexicographically."""
        return sorted(self.boundary, key=lambda s: (len(s), s))

    def get_simplices_of_dim(self, dim: int) -> list:
        return sorted(self.dim_simplex_dict.get(dim, []))

    def get_simplices_of_dim_count(self, dim: int) -> int:
        return len(self.dim_simplex_dict.get(dim, []))

    def get_sub_complex_simplices(self) -> list:
        """Return list of all simplices of the sub_complex, sorted by dimension and then lexicographically."""
        return sorted(self.sub_complex, key=lambda s: (len(s), s))

    def set_simplex_weights(self, weight_function: dict, default_value: float = 0) -> None:
        """
        Add weights/radii of simplices from a dictionary `simplex : weight`.
        Add default_value to all simplices not present in the weight function.

        Note: Monotonicity is NOT checked.
        """
        self.simplex_weights = {simplex: default_value for simplex in self.boundary}
        for simplex, weight in weight_function.items():
            simplex_tuple = tuple(sorted(simplex))
            if simplex_tuple not in self.boundary:
                raise ValueError(f"Simplex {simplex_tuple} is not in the simplicial complex, cannot add its weight.")
            if weight < default_value:
                raise Warning(f"Simplex {simplex_tuple} given weight lower than default.")
            self.simplex_weights[simplex_tuple] = weight
        co_boundary = (self.co_boundary if self.co_boundary
                       else BoundaryMatrixUtils.make_co_boundary(self.boundary))
        FloatingPointUtils.ensure_weights_monotonicity_and_equal_values(self.simplex_weights, co_boundary)

    def get_weight_function_copy(self) -> dict:
        """Return copy of {simplex : weight} dictionary."""
        return {simplex: weight for simplex, weight in self.simplex_weights.items()}

    def get_simplex_weight(self, simplex: tuple[int, ...]) -> float:
        """Return the weight of a simplex."""
        if simplex in self.simplex_weights:
            return self.simplex_weights[simplex]
        else:
            raise KeyError(f'{simplex} is not a simplex in the simplicial complex (or has no wight attached to it).')

    def get_extra_vertices_of_cofaces(self, simplex: tuple[int, ...]) -> list[int, ...]:
        return [(set(co_face) - set(simplex)).pop() for co_face in self.co_boundary[simplex]]

    def set_sub_complex(self, simplices) -> None:
        """Sets a sub_complex generated by given list of simplices."""
        simplices = set(tuple(sorted(s)) for s in simplices)
        queue = sorted(simplices, key=lambda s: (len(s), s))
        while queue:
            simplex = queue.pop()
            if simplex not in self.boundary:
                raise KeyError(f"Simplex {simplex} is not in the simplicial complex, cannot add into sub-complex.")
            simplex_boundary = self.boundary[simplex]
            for sub_simplex in simplex_boundary:
                if sub_simplex not in simplices:
                    simplices.add(sub_simplex)
                    queue.append(sub_simplex)
        self.sub_complex = simplices

    def set_total_sub_complex(self, vertices) -> None:
        """Sets a sub_complex as the total sub_complex given by a list of vertices."""
        vertices = set(vertices)
        self.sub_complex = set(s for s in self.boundary if set(s).issubset(vertices))

    def get_dimension(self) -> int:
        return max(self.dim_simplex_dict, default=-1)

    def chain_boundary(self, simplices) -> set:
        """Return boundary of a chain of simplices, i.e., the symmetric difference of boundaries of the given collection
         of simplices."""
        boundary = set()
        for simplex in simplices:
            boundary ^= self.boundary[simplex]
        return boundary


import unittest

from chromatic_tda import ChromaticAlphaComplex
from chromatic_tda.entities.simplicial_complex import SimplicialComplex


class ComplexTest(unittest.TestCase):

    def test_chromatic_subcomplex_subcomplex_mono_chromatic(self) -> None:
        simplicial_complex = SimplicialComplex([(0, 1, 2), (1, 2, 3)])

        chromatic_subcomplex = simplicial_complex.get_chromatic_subcomplex([0, 1, 1, 0], sub_complex='mono-chromatic')

        assert set(chromatic_subcomplex.simplices()) == {(0,), (1,), (2,), (3,), (0, 1), (0, 2), (1, 2), (1, 3), (2, 3),
                                                         (0, 1, 2), (1, 2, 3)}
        assert set(chromatic_subcomplex.simplices_sub_complex()) == {(0,), (3,), (1,), (2,), (1, 2)}

    def test_chromatic_subcomplex_subcomplex_all(self) -> None:
        simplicial_complex = SimplicialComplex([(0, 1, 2), (1, 2, 3)])

        chromatic_subcomplex = simplicial_complex.get_chromatic_subcomplex([0, 1, 1, 0], sub_complex='all')

        assert set(chromatic_subcomplex.simplices()) == {(0,), (1,), (2,), (3,), (0, 1), (0, 2), (1, 2), (1, 3), (2, 3),
                                                         (0, 1, 2), (1, 2, 3)}
        assert set(chromatic_subcomplex.simplices_sub_complex()) == {(0,), (1,), (2,), (3,), (0, 1), (0, 2), (1, 2), (1, 3), (2, 3),
                                                         (0, 1, 2), (1, 2, 3)}

    def test_chromatic_subcomplex_subcomplex_unused_label(self) -> None:
        simplicial_complex = SimplicialComplex([(0, 1, 2), (1, 2, 3)])

        chromatic_subcomplex = simplicial_complex.get_chromatic_subcomplex([0, 1, 1, 0], sub_complex='2',
                                                                           allow_unused_labels=True)

        assert set(chromatic_subcomplex.simplices()) == {(0,), (1,), (2,), (3,), (0, 1), (0, 2), (1, 2), (1, 3), (2, 3),
                                                         (0, 1, 2), (1, 2, 3)}
        assert set(chromatic_subcomplex.simplices_sub_complex()) == set()

    def test_chromatic_subcomplex_complex_mono_chromatic(self) -> None:
        simplicial_complex = SimplicialComplex([(0, 1, 2), (1, 2, 3)])

        chromatic_subcomplex = simplicial_complex.get_chromatic_subcomplex([0, 1, 1, 0], full_complex='mono-chromatic')

        assert set(chromatic_subcomplex.simplices()) == {(0,), (3,), (1,), (2,), (1, 2)}
        assert set(chromatic_subcomplex.simplices_sub_complex()) == set()

    def test_chromatic_subcomplex_complex_mono_chromatic_list(self) -> None:
        simplicial_complex = SimplicialComplex([(0, 1, 2), (1, 2, 3)])

        chromatic_subcomplex = simplicial_complex.get_chromatic_subcomplex([0, 1, 1, 0], full_complex=[[0], [1]])

        assert set(chromatic_subcomplex.simplices()) == {(0,), (3,), (1,), (2,), (1, 2)}
        assert set(chromatic_subcomplex.simplices_sub_complex()) == set()

    def test_chromatic_subcomplex_complex_all(self) -> None:
        simplicial_complex = SimplicialComplex([(0, 1, 2), (1, 2, 3)])

        chromatic_subcomplex = simplicial_complex.get_chromatic_subcomplex([0, 1, 1, 0], full_complex='all')

        assert set(chromatic_subcomplex.simplices()) == {(0,), (1,), (2,), (3,), (0, 1), (0, 2), (1, 2), (1, 3), (2, 3),
                                                         (0, 1, 2), (1, 2, 3)}
        assert set(chromatic_subcomplex.simplices_sub_complex()) == set()

    def test_chromatic_subcomplex_relative_mono_chromatic(self) -> None:
        simplicial_complex = SimplicialComplex([(0, 1, 2), (1, 2, 3)])
        simplicial_complex.set_sub_complex([(0, 1, 2)])

        chromatic_subcomplex = simplicial_complex.get_chromatic_subcomplex([0, 1, 1, 0], relative='mono-chromatic')

        assert set(chromatic_subcomplex.simplices()) == {(0, 1), (0, 2), (1, 3), (2, 3), (0, 1, 2), (1, 2, 3)}
        assert set(chromatic_subcomplex.simplices_sub_complex()) == set()

    def test_chromatic_subcomplex_relative_mono_chromatic_list(self) -> None:
        simplicial_complex = SimplicialComplex([(0, 1, 2), (1, 2, 3)])
        simplicial_complex.set_sub_complex([(0, 1, 2)])

        chromatic_subcomplex = simplicial_complex.get_chromatic_subcomplex([0, 1, 1, 0], relative=[[0], [1]])

        assert set(chromatic_subcomplex.simplices()) == {(0, 1), (0, 2), (1, 3), (2, 3), (0, 1, 2), (1, 2, 3)}
        assert set(chromatic_subcomplex.simplices_sub_complex()) == set()

    def test_chromatic_subcomplex_relative_all(self) -> None:
        simplicial_complex = SimplicialComplex([(0, 1, 2), (1, 2, 3)])

        chromatic_subcomplex = simplicial_complex.get_chromatic_subcomplex([0, 1, 1, 0], relative='all')

        assert set(chromatic_subcomplex.simplices()) == set()
        assert set(chromatic_subcomplex.simplices_sub_complex()) == set()

    def test_chromatic_subcomplex_bi_to_tri_over_mono(self) -> None:
        simplicial_complex = SimplicialComplex([(0, 1, 2, 3, 4), (2, 3, 5)])
        chromatic_subcomplex = simplicial_complex.get_chromatic_subcomplex([0, 1, 2, 3, 3, 2],
                                                                           sub_complex='bi-chromatic',
                                                                           full_complex='tri-chromatic',
                                                                           relative='mono-chromatic')

        assert set(chromatic_subcomplex.simplices()) == {(0, 1), (0, 2), (0, 3), (0, 4), (1, 2), (1, 3), (1, 4), (2, 3),
                                                         (2, 4), (3, 5), (0, 1, 2), (0, 1, 3), (0, 1, 4), (0, 2, 3),
                                                         (0, 2, 4), (0, 3, 4), (1, 2, 3), (1, 2, 4), (1, 3, 4),
                                                         (2, 3, 4), (2, 3, 5), (0, 1, 3, 4), (0, 2, 3, 4), (1, 2, 3, 4)}
        assert set(chromatic_subcomplex.simplices_sub_complex()) == {(0, 1), (0, 2), (0, 3), (0, 4), (1, 2), (1, 3),
                                                                     (1, 4), (2, 3), (2, 4), (3, 5), (0, 3, 4),
                                                                     (1, 3, 4), (2, 3, 4), (2, 3, 5)}

    def test_chromatic_subcomplex_custom_labels(self) -> None:
        simplicial_complex = SimplicialComplex([(0, 1, 2), (1, 2, 3)])

        chromatic_subcomplex = simplicial_complex.get_chromatic_subcomplex(['blue', 'red', 'red', 'blue'],
                                                                           sub_complex='mono-chromatic')

        assert set(chromatic_subcomplex.simplices()) == {(0,), (1,), (2,), (3,), (0, 1), (0, 2), (1, 2), (1, 3), (2, 3),
                                                         (0, 1, 2), (1, 2, 3)}
        assert set(chromatic_subcomplex.simplices_sub_complex()) == {(0,), (3,), (1,), (2,), (1, 2)}

    def test_chromatic_subcomplex_01_2(self) -> None:
        simplicial_complex = SimplicialComplex([(0, 1, 2), (1, 2, 3), (2, 3, 4), (1, 3, 5), (4, 5)])

        chromatic_subcomplex = simplicial_complex.get_chromatic_subcomplex([0, 1, 1, 0, 2, 2],
                                                                           sub_complex='01,2')
        assert set(chromatic_subcomplex.simplices_sub_complex()) == {(0,), (1,), (2,), (3,), (4,), (5,), (0, 1), (0, 2),
                                                                     (1, 2), (1, 3), (2, 3), (4, 5), (0, 1, 2),
                                                                     (1, 2, 3)}

    def test_chromatic_subcomplex_ab_c(self) -> None:
        simplicial_complex = SimplicialComplex([(0, 1, 2), (1, 2, 3), (2, 3, 4), (1, 3, 5), (4, 5)])

        chromatic_subcomplex = simplicial_complex.get_chromatic_subcomplex(['a', 'b', 'b', 'a', 'c', 'c'],
                                                                           sub_complex='ab,c')
        assert set(chromatic_subcomplex.simplices_sub_complex()) == {(0,), (1,), (2,), (3,), (4,), (5,), (0, 1), (0, 2),
                                                                     (1, 2), (1, 3), (2, 3), (4, 5), (0, 1, 2),
                                                                     (1, 2, 3)}

    def test_chromatic_subcomplex_ab_c_list(self) -> None:
        simplicial_complex = SimplicialComplex([(0, 1, 2), (1, 2, 3), (2, 3, 4), (1, 3, 5), (4, 5)])

        chromatic_subcomplex = simplicial_complex.get_chromatic_subcomplex(['a', 'b', 'b', 'a', 'c', 'c'],
                                                                           sub_complex=[['a', 'b'], ['c']])
        assert set(chromatic_subcomplex.simplices_sub_complex()) == {(0,), (1,), (2,), (3,), (4,), (5,), (0, 1), (0, 2),
                                                                     (1, 2), (1, 3), (2, 3), (4, 5), (0, 1, 2),
                                                                     (1, 2, 3)}

    def test_chromatic_subcomplex_dictionary_labels(self) -> None:
        simplicial_complex = SimplicialComplex([(0, 1, 2), (1, 2, 3)])

        chromatic_subcomplex = simplicial_complex.get_chromatic_subcomplex({
            0: '0', 1: '1', 2: '1', 3: '0'
        }, sub_complex='mono-chromatic')

        assert set(chromatic_subcomplex.simplices()) == {(0,), (1,), (2,), (3,), (0, 1), (0, 2), (1, 2), (1, 3), (2, 3),
                                                         (0, 1, 2), (1, 2, 3)}
        assert set(chromatic_subcomplex.simplices_sub_complex()) == {(0,), (3,), (1,), (2,), (1, 2)}

    def test_chromatic_subcomplex_alpha_unused_label(self) -> None:
        points = [[0, 0], [1.1, 1.2], [.9, -.1], [.3, .4], [.5, .7], [.123, .432]]
        labels = [1, 1, 1, 1, 1, 1]
        simplicial_complex = ChromaticAlphaComplex(points, labels).get_simplicial_complex(sub_complex='0',
                                                                                          allow_unused_labels=True)

        assert len(simplicial_complex.simplices()) > 0
        assert len(simplicial_complex.simplices_sub_complex()) == 0



from time import perf_counter

from chromatic_tda.utils.singleton import singleton


@singleton
class TimingUtils:
    total_time_dict : dict[str, float]
    started_at_dict : dict[str, float]
    call_count_dict : dict[str, int]
    log_times : bool

    def __init__(self, log_times=False):
        self.log_times = log_times
        self.flush()

    def flush(self):
        self.total_time_dict = {}
        self.started_at_dict = {}
        self.call_count_dict = {}

    def start(self, topic: str):
        self.started_at_dict[topic] = perf_counter()
        if self.log_times:
            if topic in self.call_count_dict:
                self.call_count_dict[topic] += 1
            else:
                self.call_count_dict[topic] = 1

    def stop(self, topic: str):
        if self.log_times:
            if topic in self.total_time_dict:
                self.total_time_dict[topic] += perf_counter() - self.started_at_dict[topic]
            else:
                self.total_time_dict[topic] = perf_counter() - self.started_at_dict[topic]

    def print(self):
        for topic in sorted(self.total_time_dict):
            print(f"Time[{topic:<58}] = {self.total_time_dict[topic]:6.2f} s "
                  f"(# calls : {self.call_count_dict[topic]:<6}, "
                  f"average = {self.total_time_dict[topic]/self.call_count_dict[topic]:7.4f})")




class MatrixReduction:
    @staticmethod
    def reduce(matrix, order_function, order_function_row = None, return_reduction_matrix = False):
        """
        Reduce given matrix. Using `order_function` to order columns.
        If `order_function_row` is given, it is used to order rows,
        otherwise `order_function` is also used for rows.
        The matrix is copied, i.e., given `matrix` remains unchanged.

        Returns a dictionary with the following keys:
            reduced_matrix ... the reduced boundary matrix
            pivots ... pivots of the reduced matrix as a dictionary: row indices as keys and column indices as values
            reduction_matrix (if return_reduction_matrix=True) ... the matrix V s.t. reduced_matrix = matrix * V
        """
        if order_function_row is None:
            order_function_row = order_function
        R = {k : v for k, v in matrix.items()}
        if return_reduction_matrix:
            V = {k : {k} for k in matrix}
        low_inv = {}  # low_inv[i]=index of column with the lowest 1 at i
        for s in sorted(R, key=order_function):
            t = low_inv.get(max(R[s], key=order_function_row), -1) if len(R[s]) != 0 else -1
            while t != -1:
                R[s] = R[t] ^ R[s]  # symmetric difference of t-th and s-th columns
                if return_reduction_matrix:
                    V[s] = V[t] ^ V[s]
                t = low_inv.get(max(R[s], key=order_function_row), -1) if len(R[s]) != 0 else -1
            if len(R[s]) != 0:
                low_inv[max(R[s], key=order_function_row)] = s

        return_dictionary = {'reduced_matrix': R, 'pivots': low_inv}
        if return_reduction_matrix:
            return_dictionary['reduction_matrix'] = V
        return return_dictionary



import unittest
import numpy as np

from chromatic_tda.entities.simplicial_complex import SimplicialComplex


class ComplexTest(unittest.TestCase):

    def test_custom_complex_without_weights_list(self) -> None:
        simplicial_complex = SimplicialComplex([(0, 2, 1), (3, 1, 2)])  # out of order to check sorting

        assert set(simplicial_complex.simplices()) == {(0,), (1,), (2,), (3,), (0, 1), (0, 2), (1, 2), (1, 3), (2, 3),
                                                       (0, 1, 2), (1, 2, 3)}
        assert set(simplicial_complex.simplices_of_dim(0)) == {(0,), (1,), (2,), (3,)}
        assert set(simplicial_complex.simplices_of_dim(1)) == {(0, 1), (0, 2), (1, 2), (1, 3), (2, 3)}
        assert set(simplicial_complex.simplices_of_dim(2)) == {(0, 1, 2), (1, 2, 3)}

    def test_custom_complex_without_weights_ndarray(self) -> None:
        simplicial_complex = SimplicialComplex(np.array([(0, 2, 1), (3, 1, 2)]))  # out of order to check sorting

        assert set(simplicial_complex.simplices()) == {(0,), (1,), (2,), (3,), (0, 1), (0, 2), (1, 2), (1, 3), (2, 3),
                                                       (0, 1, 2), (1, 2, 3)}
        assert set(simplicial_complex.simplices_of_dim(0)) == {(0,), (1,), (2,), (3,)}
        assert set(simplicial_complex.simplices_of_dim(1)) == {(0, 1), (0, 2), (1, 2), (1, 3), (2, 3)}
        assert set(simplicial_complex.simplices_of_dim(2)) == {(0, 1, 2), (1, 2, 3)}

    def test_custom_complex_with_weights(self) -> None:
        simplicial_complex = SimplicialComplex({
            (1, 2): 1,
            (3,): 2,
            (1, 3): 2,
            (2, 3): 2,
            (1, 2, 3): 3,
            (0, 1, 2): 4
        })

        assert simplicial_complex.weight_function() == {(0, 1, 2): 4,
                                                        (1, 2, 3): 3,
                                                        (0, 1): 0,
                                                        (2, 3): 2,
                                                        (0, 2): 0,
                                                        (1, 2): 1,
                                                        (1, 3): 2,
                                                        (1,): 0,
                                                        (2,): 0,
                                                        (0,): 0,
                                                        (3,): 2}



import numpy as np
import multiprocessing
import math

from chromatic_tda.utils.singleton import singleton
from chromatic_tda.core.core_chromatic_alpha_complex import CoreChromaticAlphaComplex
from chromatic_tda.utils.legacy_geometrical_utils import intersect_lines, project_to_line, bisector, \
    circum_center, sq_dist, sq_dist_max
from chromatic_tda.core.core_simplicial_complex import CoreSimplicialComplex
from chromatic_tda.utils.timing import TimingUtils


# class RadiusFunctionParallelUtils:
#
#     atoms : int
#     alpha_complex : CoreChromaticAlphaComplex
#     simplex_list : list
#
#     def __init__(self, alpha_complex) -> None:
#         # we can use os.cpu_count()
#         self.atoms = 5
#         self.alpha_complex = alpha_complex
#
#     def calculate_radius_function(self, simplex):
#         dim = len(simplex) - 1
#
#         if dim == 4:
#             return RadiusFunctionUtils._compute_radius_function_pentachoron(self.alpha_complex, simplex)
#         elif dim == 3:
#             return RadiusFunctionUtils._compute_radius_function_tetrahedron(self.alpha_complex, simplex)
#         elif dim == 2:
#             return RadiusFunctionUtils._compute_radius_function_triangle(self.alpha_complex, simplex)
#         elif dim == 1:
#             return RadiusFunctionUtils._compute_radius_function_edge(self.alpha_complex, simplex)
#
#         raise NotImplementedError()
#
#     def calculate(self, atom):
#         actions = math.ceil(len(self.simplex_list)/self.atoms)
#         res = []
#         for i in range(0, actions):
#             k = atom * actions + i
#             if k >= len(self.simplex_list):
#                 break
#             simplex = self.simplex_list[k]
#             rad = self.calculate_radius_function(simplex)
#             res.append((simplex, rad))
#         return res
#
#     def compute_radius_function_in_parallel(self, **kwargs) -> None:
#         TimingUtils().start("Compute Radius Function Parallel")
#         print("Parallel Processing Started")
#
#         for dim in [4, 3, 2, 1]:
#             size_of_process = len(self.alpha_complex.simplicial_complex.dim_simplex_dict.get(dim, {}))
#
#             if size_of_process == 0:
#                 continue
#
#             self.simplex_list = []
#             for simplex in self.alpha_complex.simplicial_complex.dim_simplex_dict.get(dim, {}):
#                 self.simplex_list.append(simplex)
#
#             with multiprocessing.Pool() as pool:
#                 res_list = pool.map(self.calculate, range(0, self.atoms))
#
#                 for res in res_list:
#                     for item in res:
#                         self.alpha_complex.sq_rad[item[0]] = item[1]
#
#         if 'round' in kwargs:
#             self.alpha_complex.simplicial_complex.set_simplex_weights(
#                 {s : np.round(np.sqrt(r), decimals=kwargs['round']) for s,r in self.alpha_complex.sq_rad.items()},
#                 default_value=0)
#         else:
#             self.alpha_complex.simplicial_complex.set_simplex_weights({s : np.sqrt(r) for s,r in self.alpha_complex.sq_rad.items()}, default_value = 0)
#
#         TimingUtils().stop("Compute Radius Function Parallel")


class LegacyRadiusFunctionUtils:

    @staticmethod
    def compute_radius_function(alpha_complex: CoreChromaticAlphaComplex, **kwargs) -> None:
        TimingUtils().start("Compute Radius Function")

        alpha_complex.sq_rad = {}

        for simplex in alpha_complex.simplicial_complex.dim_simplex_dict.get(4, {}):
            alpha_complex.sq_rad[simplex] = LegacyRadiusFunctionUtils._compute_radius_function_pentachoron(alpha_complex, simplex)
        for simplex in alpha_complex.simplicial_complex.dim_simplex_dict.get(3, {}):
            alpha_complex.sq_rad[simplex] = LegacyRadiusFunctionUtils._compute_radius_function_tetrahedron(alpha_complex, simplex)
        for simplex in alpha_complex.simplicial_complex.dim_simplex_dict.get(2, {}):
            alpha_complex.sq_rad[simplex] = LegacyRadiusFunctionUtils._compute_radius_function_triangle(alpha_complex, simplex)
        for simplex in alpha_complex.simplicial_complex.dim_simplex_dict.get(1, {}):
            alpha_complex.sq_rad[simplex] = LegacyRadiusFunctionUtils._compute_radius_function_edge(alpha_complex, simplex)
        if 'round' in kwargs:
            alpha_complex.simplicial_complex.set_simplex_weights(
                {s : np.round(np.sqrt(r), decimals=kwargs['round']) for s,r in alpha_complex.sq_rad.items()},
                default_value=0)
        else:
            alpha_complex.simplicial_complex.set_simplex_weights({s : np.sqrt(r) for s,r in alpha_complex.sq_rad.items()}, default_value = 0)

        TimingUtils().stop("Compute Radius Function")

    @staticmethod
    def _compute_radius_function_pentachoron(alpha_complex: CoreChromaticAlphaComplex, simplex):
        TimingUtils().start("Compute Radius Function :: Pentachoron")

        red, grn, blu = (alpha_complex.points[list(s)] for s in alpha_complex.OLD_split_simplex_sort_by_size(simplex))
        if (len(red), len(grn), len(blu)) == (2,2,1):
            E_pt = intersect_lines(*bisector(*red), *bisector(*grn))
            res = sq_dist_max(E_pt, red[0], grn[0], blu[0])
        elif (len(red), len(grn), len(blu)) == (3,1,1):
            E_pt = circum_center(*red)
            res = sq_dist_max(E_pt, red[0], grn[0], blu[0])
        else:
            raise RuntimeError(f"Simplex colored {len(red)}:{len(grn)}:{len(blu)} cannot occur in 2D.")

        TimingUtils().stop("Compute Radius Function :: Pentachoron")
        return res

    @staticmethod
    def _compute_radius_function_tetrahedron(alpha_complex: CoreChromaticAlphaComplex, simplex):
        TimingUtils().start("Compute Radius Function :: Tetrahedron")

        red, grn, blu = (alpha_complex.points[list(s)] for s in alpha_complex.OLD_split_simplex_sort_by_size(simplex))
        if (len(red), len(grn), len(blu)) == (3,1,0):
            E_pt = circum_center(*red)
            if alpha_complex.is_empty_stack(E_pt, simplex):
                res = sq_dist_max(E_pt, red[0], grn[0])
            else:
                res = min(alpha_complex.sq_rad[cf] for cf in alpha_complex.simplicial_complex.co_boundary[simplex])

        elif (len(red), len(grn), len(blu)) == (2,2,0):
            E_pt = intersect_lines(*bisector(*red), *bisector(*grn))
            if alpha_complex.is_empty_stack(E_pt, simplex):
                res = sq_dist_max(E_pt, red[0], grn[0])
            else:
                res = min(alpha_complex.sq_rad[cf] for cf in alpha_complex.simplicial_complex.co_boundary[simplex])

        elif (len(red), len(grn), len(blu)) == (2,1,1):
            p, p_dir = bisector(*red)
            pts = [
                project_to_line(p, p_dir, red[0]),
                project_to_line(p, p_dir, grn[0]),
                project_to_line(p, p_dir, blu[0]),
                intersect_lines(p, p_dir, *bisector(red[0], grn[0])),
                intersect_lines(p, p_dir, *bisector(red[0], blu[0])),
                intersect_lines(p, p_dir, *bisector(grn[0], blu[0]))
            ]
            radii = [max(sq_dist(pt, red[0]),
                         sq_dist(pt, grn[0]),
                         sq_dist(pt, blu[0])) for pt in pts]

            radius, center_index = min((rad, ind) for ind, rad in enumerate(radii))
            center = pts[center_index]
            if alpha_complex.is_empty_stack(center, simplex):
                res = radius
            else:
                res = min(alpha_complex.sq_rad[cf] for cf in alpha_complex.simplicial_complex.co_boundary[simplex])
        else:
            raise RuntimeError(f"Simplex colored {len(red)}:{len(grn)}:{len(blu)} cannot occur in 2D.")

        TimingUtils().stop("Compute Radius Function :: Tetrahedron")
        return res

    @staticmethod
    def _compute_radius_function_triangle(alpha_complex: CoreChromaticAlphaComplex, simplex):
        TimingUtils().start("Compute Radius Function :: Triangle")

        red, grn, blu = (alpha_complex.points[list(s)] for s in alpha_complex.OLD_split_simplex_sort_by_size(simplex))

        if (len(red), len(grn), len(blu)) == (3,0,0):
            E_pt = circum_center(*red)
            if alpha_complex.is_empty_stack(E_pt, simplex):
                res = sq_dist_max(E_pt, red[0])
            else:
                res = min(alpha_complex.sq_rad[cf] for cf in alpha_complex.simplicial_complex.co_boundary[simplex])

        elif (len(red), len(grn), len(blu)) == (2,1,0):
            p, p_dir = bisector(*red)
            
            pts = [
                project_to_line(p, p_dir, red[0]),
                project_to_line(p, p_dir, grn[0]),
                intersect_lines(p, p_dir, *bisector(red[0], grn[0]))
            ]
            radii = [max(sq_dist(pt, red[0]),
                         sq_dist(pt, grn[0])) for pt in pts]

            radius, center_index = min((rad, ind) for ind, rad in enumerate(radii))
            center = pts[center_index]
            if alpha_complex.is_empty_stack(center, simplex):
                res = radius
            else:
                res = min(alpha_complex.sq_rad[cf] for cf in alpha_complex.simplicial_complex.co_boundary[simplex])

        elif (len(red), len(grn), len(blu)) == (1,1,1):
            pts = [ # no need to check red[0],grn[0],blu[0], max rad is never maxed there
                bisector(red[0], grn[0])[0],
                bisector(red[0], blu[0])[0],
                bisector(grn[0], blu[0])[0],
                circum_center(red[0], grn[0], blu[0])
            ]
            radii = [max(sq_dist(pt, red[0]),
                         sq_dist(pt, grn[0]),
                         sq_dist(pt, blu[0])) for pt in pts]
            radius, center_index = min((rad, ind) for ind, rad in enumerate(radii))
            center = pts[center_index]
            if alpha_complex.is_empty_stack(center, simplex):
                res = radius
            else:
                res = min(alpha_complex.sq_rad[cf] for cf in alpha_complex.simplicial_complex.co_boundary[simplex])
        else:
            raise RuntimeError(f"Simplex colored {len(red)}:{len(grn)}:{len(blu)} cannot occur in 2D.")

        TimingUtils().stop("Compute Radius Function :: Triangle")
        return res

    @staticmethod
    def _compute_radius_function_edge(alpha_complex: CoreChromaticAlphaComplex, simplex):
        TimingUtils().start("Compute Radius Function :: Edge")

        red, grn, blu = (alpha_complex.points[list(s)] for s in alpha_complex.OLD_split_simplex_sort_by_size(simplex))
        if (len(red), len(grn), len(blu)) == (2,0,0):
            E_pt, _ = bisector(*red)
            if alpha_complex.is_empty_stack(E_pt, simplex):
                res = sq_dist_max(E_pt, red[0])
            else:
                res = min(alpha_complex.sq_rad[cf] for cf in alpha_complex.simplicial_complex.co_boundary[simplex])
        elif (len(red), len(grn), len(blu)) == (1,1,0):
            E_pt, _ = bisector(red[0], grn[0])
            if alpha_complex.is_empty_stack(E_pt, simplex):
                res = sq_dist_max(E_pt, red[0], grn[0]) #grn[0] not needed here
            else:
                res = min(alpha_complex.sq_rad[cf] for cf in alpha_complex.simplicial_complex.co_boundary[simplex])
        else:
            raise RuntimeError(f"Simplex colored {len(red)}:{len(grn)}:{len(blu)} cannot occur in 2D.")

        TimingUtils().stop("Compute Radius Function :: Edge")
        return res

    @staticmethod
    def check_monotonicity_of_radius_function(complex: CoreSimplicialComplex):
        return all(all(complex.simplex_weights[b] <= complex.simplex_weights[simplex] for b in boundary)
                   for simplex, boundary in complex.boundary.items())



import numpy as np
import numpy.typing as npt

from chromatic_tda.algorithms.chromatic_subcomplex_utils import ChromaticComplexUtils
from chromatic_tda.core.core_simplicial_complex import CoreSimplicialComplex
from chromatic_tda.utils.legacy_geometrical_utils import sq_dist


class CoreChromaticAlphaComplex:
    points: npt.NDArray
    points_dimension: int
    input_labels_to_internal_labels_dict: dict
    internal_labels_to_input_labels_dict: dict
    labels_number: int
    internal_labeling: list
    simplicial_complex: CoreSimplicialComplex

    def __init__(self) -> None:
        self.input_labels_to_internal_labels_dict = {}
        self.labels_number = 0
        self.internal_labeling = []
        self.sq_rad = {}

    def __iter__(self):
        yield from self.simplicial_complex

    def __len__(self) -> int:
        return len(self.simplicial_complex)

    def __contains__(self, element) -> bool:
        return element in self.simplicial_complex

    def copy_points(self):
        """Return a copy of the points"""
        return np.array(self.points)

    def simplex_labels_internal(self, simplex):
        return {self.internal_labeling[v] for v in simplex}

    def simplex_labels_input(self, simplex):
        """Return set of labels of the vertices of the given simplex"""
        return {self.internal_labels_to_input_labels_dict[lab] for lab in self.simplex_labels_internal(simplex)}

    def get_simplicial_complex(self, sub_complex, full_complex, relative, allow_unused_labels) -> CoreSimplicialComplex:
        return ChromaticComplexUtils.get_chromatic_subcomplex(
            sub_complex=sub_complex, full_complex=full_complex, relative=relative,
            simplicial_complex=self.simplicial_complex, internal_labeling=self.internal_labeling,
            labels_user_to_internal=self.input_labels_to_internal_labels_dict,
            allow_unused_labels=allow_unused_labels
        )

    def star_vertices(self, simplex) -> set:
        return set().union(*self.simplicial_complex.co_boundary[simplex]) - set(simplex)

    def is_empty_stack(self, center, simplex) -> bool:
        """
        Return True if the stack centered at `center` is empty.
        Assume that the center given is a circum-center for each
        monochromatic face, and check whether the vertices in
        co-boundary of simplex are inside the circum-centers of
        the corresponding colors. Colors not present in simplex
        are treated as automatically satisfied.
        """
        vertices_to_check = self.star_vertices(simplex)
        for mono_chrom_face in self.OLD_split_simplex(simplex):  # for every color
            if len(mono_chrom_face) > 0:  # if the color is present in simplex
                radius = sq_dist(center, self.points[mono_chrom_face[0]])
                if any((self.internal_labeling[mono_chrom_face[0]] == self.internal_labeling[v] and
                        sq_dist(center, self.points[v]) < radius)
                       for v in vertices_to_check):
                    return False
        return True

    def OLD_split_simplex(self, simplex) -> list:
        return [tuple(i for i in simplex if self.internal_labeling[i] == 0),
                tuple(i for i in simplex if self.internal_labeling[i] == 1),
                tuple(i for i in simplex if self.internal_labeling[i] == 2)]

    def OLD_split_simplex_sort_by_size(self, simplex) -> list:
        return sorted(self.OLD_split_simplex(simplex), key=len, reverse=True)

    def write(self) -> None:
        print()
        print(f"**** Delaunay Complex (dimension = {self.simplicial_complex.get_dimension()}) ****")
        for (p, l) in zip(self.points, self.internal_labeling):
            print(f"{p} -> {l}")
        print(42 * "*")
        print()
